epan_tvbuff_h_types_cdef = """


/* tvbuff.h
 *
 * Testy, Virtual(-izable) Buffer of guint8*'s
 *
 * "Testy" -- the buffer gets mad when an attempt is made to access data
 *      beyond the bounds of the buffer. An exception is thrown.
 *
 * "Virtual" -- the buffer can have its own data, can use a subset of
 *      the data of a backing tvbuff, or can be a composite of
 *      other tvbuffs.
 *
 * Copyright (c) 2000 by Gilbert Ramirez <gram@alumni.rice.edu>
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/**
 * "testy, virtual(-izable) buffer".  They are testy in that they get mad when
 * an attempt is made to access data beyond the bounds of their array. In that
 * case, they throw an exception.
 *
 * They are virtualizable in that new tvbuff's can be made from other tvbuffs,
 * while only the original tvbuff may have data. That is, the new tvbuff has
 * virtual data.
 */

struct tvbuff;
typedef struct tvbuff tvbuff_t;

/**
 * Given a tvbuff, an offset into the tvbuff, and a length that starts
 * at that offset (which may be -1 for "all the way to the end of the
 * tvbuff"), fetch BCD encoded digits from a tvbuff starting from either
 * the low or high half byte, formatting the digits according to an input digit
 * set, if NUL a default digit set of 0-9 returning "?" for overdecadic digits
 * will be used.  A pointer to the packet-scope (WMEM-allocated) string will
 * be returned. Note a tvbuff content of 0xf is considered a 'filler' and will
 * end the conversion.
 */
typedef struct dgt_set_t
{
    const unsigned char out[16];
}
dgt_set_t;

typedef void (*tvbuff_free_cb_t)(void*);
"""

epan_tvbuff_h_funcs_cdef = """
/** @defgroup tvbuff Testy, Virtual(-izable) Buffers
 *
 * Dissector use and management
 *
 *  Consider a collection of tvbs as being a chain or stack of tvbs.
 *
 *  When dissecting a frame:
 *   The top-level dissector (packet.c) pushes the initial tvb (containing
 *   the complete frame) onto the stack (starts the chain) and then calls
 *   a sub-dissector which in turn calls the next sub-dissector and so on.
 *   Each sub-dissector may chain additional tvbs (see below) to the tvb
 *   handed to that dissector. After dissection is complete and control has
 *   returned to the top-level dissector, the chain of tvbs (stack) is free'd
 *   via a call to tvb_free_chain() (in epan_dissect_cleanup()).
 *
 * A dissector:
 *  - Can chain new tvbs (subset, real, composite) to the
 *    tvb handed to the dissector using tvb_new_subset_length_caplen(),
 *    tvb_new_subset_length(), tvb_new_subset_remaining(),
 *    tvb_new_child_real_data(), tvb_set_child_real_data_tvbuff(),
 *    tvb_composite_finalize(), and tvb_child_uncompress(). (Composite
 *    tvbs should reference only tvbs which are already part of the chain).
 *  - Must not save for later use (e.g., when dissecting another frame) a
 *    pointer to a tvb handed to the dissector; (A higher level function
 *    may very well free the chain thus leaving a dangling pointer).
 *    This (obviously) also applies to any tvbs chained to the tvb handed
 *    to the dissector.
 *  - Can create its own tvb chain (using tvb_new_real_data() which the
 *
 *    dissector is free to manage as desired.
 * @{
 */

/** A "real" tvbuff contains a guint8* that points to real data.
 * The data is allocated and contiguous.
 *
 * A "subset" tvbuff has a backing tvbuff. It is a "window" through
 * which the program sees only a portion of the backing tvbuff.
 *
 * A "composite" tvbuff combines multiple tvbuffs sequentially to
 * produce a larger byte array.
 *
 * tvbuff's of any type can be used as the backing-tvbuff of a
 * "subset" tvbuff or as a member of a "composite" tvbuff.
 * "composite" tvbuffs can have member-tvbuffs of different types.
 *
 * Once a tvbuff is create/initialized/finalized, the tvbuff is read-only.
 * That is, it cannot point to any other data. A new tvbuff must be created if
 * you want a tvbuff that points to other data.
 *
 * tvbuff's are normally chained together to allow efficient de-allocation of
 * tvbuff's.
 */

/** Extracts 'number of bits' starting at 'bit offset'.
 * Returns a pointer to a newly initialized g_malloc'd REAL_DATA
 * tvbuff with the bits octet aligned.
 */
extern tvbuff_t *tvb_new_octet_aligned(tvbuff_t *tvb,
    guint32 bit_offset, gint32 no_of_bits);

extern tvbuff_t *tvb_new_chain(tvbuff_t *parent, tvbuff_t *backing);

extern tvbuff_t *tvb_clone(tvbuff_t *tvb);

extern tvbuff_t *tvb_clone_offset_len(tvbuff_t *tvb, guint offset,
    guint len);

/** Free a tvbuff_t and all tvbuffs chained from it
 * The tvbuff must be 'the 'head' (initial) tvb of a chain or
 * must not be in a chain.
 * If specified, a callback to free the tvbuff data will be invoked
 * for each tvbuff free'd */
extern void tvb_free(tvbuff_t *tvb);

/** Free the tvbuff_t and all tvbuffs chained from it.
 * The tvbuff must be 'the 'head' (initial) tvb of a chain or
 * must not be in a chain.
 * If specified, a callback to free the tvbuff data will be invoked
 * for each tvbuff free'd */
extern void tvb_free_chain(tvbuff_t *tvb);

/** Set a callback function to call when a tvbuff is actually freed
 * One argument is passed to that callback --- a void* that points
 * to the real data. Obviously, this only applies to a
 * "real" tvbuff. */
extern void tvb_set_free_cb(tvbuff_t *tvb, const tvbuff_free_cb_t func);

/** Attach a "real" tvbuff to a parent tvbuff. This connection is used
 * during a tvb_free_chain()... the "child" "real" tvbuff acts as if it
 * is part of the chain-of-creation of the parent tvbuff, although it
 * isn't. This is useful if you need to take the data from some tvbuff,
 * run some operation on it, like decryption or decompression, and make
 * a new tvbuff from it, yet want the new tvbuff to be part of the chain.
 * The reality is that the new tvbuff *is* part of the "chain of creation",
 * but in a way that these tvbuff routines are ignorant of. Use this
 * function to make the tvbuff routines knowledgable of this fact. */
extern void tvb_set_child_real_data_tvbuff(tvbuff_t *parent,
    tvbuff_t *child);

extern tvbuff_t *tvb_new_child_real_data(tvbuff_t *parent,
    const guint8 *data, const guint length, const gint reported_length);

/** Create a tvbuff backed by existing data. Can throw ReportedBoundsError.
 * Normally, a callback to free the data should be registered using
 * tvb_set_free_cb(); when this tvbuff is freed, then your callback will be
 * called, and at that time you can free your original data. */
extern tvbuff_t *tvb_new_real_data(const guint8 *data,
    const guint length, const gint reported_length);

/** Create a tvbuff that's a subset of another tvbuff.
 *
 * 'backing_offset', if positive, is the offset from the beginning of
 * the backing tvbuff at which the new tvbuff's data begins, and, if
 * negative, is the offset from the end of the backing tvbuff at which
 * the new tvbuff's data begins.
 *
 * 'backing_length' is the length of the data to include in the new
 * tvbuff, starting with the byte at 'backing_offset"; if -1, it
 * means "to the end of the backing tvbuff".  It can be 0, although
 * the usefulness of the buffer would be rather limited.
 *
 * Will throw BoundsError if 'backing_offset'/'length'
 * is beyond the bounds of the backing tvbuff.
 * Can throw ReportedBoundsError. */
extern tvbuff_t *tvb_new_subset_length_caplen(tvbuff_t *backing,
    const gint backing_offset, const gint backing_length,
    const gint reported_length);

/**
 * Similar to tvb_new_subset_length_caplen() but with captured length calculated
 * to fit within the existing captured length and the specified
 * reported length.
 * Can throw ReportedBoundsError. */
extern tvbuff_t *tvb_new_subset_length(tvbuff_t *backing,
    const gint backing_offset, const gint reported_length);

/** Similar to tvb_new_subset_length_caplen() but with backing_length and reported_length set
 * to -1.  Can throw ReportedBoundsError. */
extern tvbuff_t *tvb_new_subset_remaining(tvbuff_t *backing,
    const gint backing_offset);

/*
* Both tvb_composite_append and tvb_composite_prepend can throw
 * BoundsError if member_offset/member_length goes beyond bounds of
 * the 'member' tvbuff. */

/** Append to the list of tvbuffs that make up this composite tvbuff */
extern void tvb_composite_append(tvbuff_t *tvb, tvbuff_t *member);

/** Prepend to the list of tvbuffs that make up this composite tvbuff */
// FIXME extern void tvb_composite_prepend(tvbuff_t *tvb, tvbuff_t *member);

/** Create an empty composite tvbuff. */
extern tvbuff_t *tvb_new_composite(void);

/** Mark a composite tvbuff as initialized. No further appends or prepends
 * occur, data access can finally happen after this finalization. */
extern void tvb_composite_finalize(tvbuff_t *tvb);


/* Get amount of captured data in the buffer (which is *NOT* necessarily the
 * length of the packet). You probably want tvb_reported_length instead. */
extern guint tvb_captured_length(const tvbuff_t *tvb);

/** Computes bytes to end of buffer, from offset (which can be negative,
 * to indicate bytes from end of buffer). Function returns 0 if offset is
 * either at the end of the buffer or out of bounds. No exception is thrown.
 * You probably want tvb_reported_length_remaining instead. */
extern gint tvb_captured_length_remaining(const tvbuff_t *tvb, const gint offset);

/** Same as above, but throws an exception if the offset is out of bounds. */
extern guint tvb_ensure_captured_length_remaining(const tvbuff_t *tvb,
    const gint offset);

/* Checks (w/o throwing exception) that the bytes referred to by
 * 'offset'/'length' actually exist in the buffer */
extern gboolean tvb_bytes_exist(const tvbuff_t *tvb, const gint offset,
    const gint length);

/** Checks that the bytes referred to by 'offset'/'length', where 'length'
 * is a 64-bit unsigned integer, actually exist in the buffer, and throws
 * an exception if they aren't. */
extern void tvb_ensure_bytes_exist64(const tvbuff_t *tvb,
    const gint offset, const guint64 length);

/** Checks that the bytes referred to by 'offset'/'length' actually exist
 * in the buffer, and throws an exception if they aren't. */
extern void tvb_ensure_bytes_exist(const tvbuff_t *tvb,
    const gint offset, const gint length);

/* Checks (w/o throwing exception) that offset exists in buffer */
extern gboolean tvb_offset_exists(const tvbuff_t *tvb,
    const gint offset);

/* Get reported length of buffer */
extern guint tvb_reported_length(const tvbuff_t *tvb);

/** Computes bytes of reported packet data to end of buffer, from offset
 * (which can be negative, to indicate bytes from end of buffer). Function
 * returns 0 if offset is either at the end of the buffer or out of bounds.
 * No exception is thrown. */
extern gint tvb_reported_length_remaining(const tvbuff_t *tvb,
    const gint offset);

/** Set the reported length of a tvbuff to a given value; used for protocols
   whose headers contain an explicit length and where the calling
   dissector's payload may include padding as well as the packet for
   this protocol.

   Also adjusts the data length. */
extern void tvb_set_reported_length(tvbuff_t *tvb, const guint);

extern guint tvb_offset_from_real_beginning(const tvbuff_t *tvb);

/* Returns the offset from the first byte of real data. */
extern gint tvb_raw_offset(tvbuff_t *tvb);

/** Set the "this is a fragment" flag. */
extern void tvb_set_fragment(tvbuff_t *tvb);

extern struct tvbuff *tvb_get_ds_tvb(tvbuff_t *tvb);


/************** START OF ACCESSORS ****************/
/* All accessors will throw an exception if appropriate */

extern guint8 tvb_get_guint8(tvbuff_t *tvb, const gint offset);
extern gint8 tvb_get_gint8(tvbuff_t *tvb, const gint offset);

extern guint16 tvb_get_ntohs(tvbuff_t *tvb, const gint offset);
extern gint16 tvb_get_ntohis(tvbuff_t *tvb, const gint offset);
extern guint32 tvb_get_ntoh24(tvbuff_t *tvb, const gint offset);
extern gint32 tvb_get_ntohi24(tvbuff_t *tvb, const gint offset);
extern guint32 tvb_get_ntohl(tvbuff_t *tvb, const gint offset);
extern gint32 tvb_get_ntohil(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_ntoh40(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_ntohi40(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_ntoh48(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_ntohi48(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_ntoh56(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_ntohi56(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_ntoh64(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_ntohi64(tvbuff_t *tvb, const gint offset);
extern gfloat tvb_get_ntohieee_float(tvbuff_t *tvb, const gint offset);
extern gdouble tvb_get_ntohieee_double(tvbuff_t *tvb,
    const gint offset);

extern guint16 tvb_get_letohs(tvbuff_t *tvb, const gint offset);
extern gint16 tvb_get_letohis(tvbuff_t *tvb, const gint offset);
extern guint32 tvb_get_letoh24(tvbuff_t *tvb, const gint offset);
extern gint32 tvb_get_letohi24(tvbuff_t *tvb, const gint offset);
extern guint32 tvb_get_letohl(tvbuff_t *tvb, const gint offset);
extern gint32 tvb_get_letohil(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_letoh40(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_letohi40(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_letoh48(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_letohi48(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_letoh56(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_letohi56(tvbuff_t *tvb, const gint offset);
extern guint64 tvb_get_letoh64(tvbuff_t *tvb, const gint offset);
extern gint64 tvb_get_letohi64(tvbuff_t *tvb, const gint offset);
extern gfloat tvb_get_letohieee_float(tvbuff_t *tvb, const gint offset);
extern gdouble tvb_get_letohieee_double(tvbuff_t *tvb,
    const gint offset);

extern guint16 tvb_get_guint16(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gint16 tvb_get_gint16(tvbuff_t *tvb, const gint offset, const guint encoding);
extern guint32 tvb_get_guint24(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gint32 tvb_get_gint24(tvbuff_t *tvb, const gint offset, const guint encoding);
extern guint32 tvb_get_guint32(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gint32 tvb_get_gint32(tvbuff_t *tvb, const gint offset, const guint encoding);
extern guint64 tvb_get_guint40(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gint64 tvb_get_gint40(tvbuff_t *tvb, const gint offset, const guint encoding);
extern guint64 tvb_get_guint48(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gint64 tvb_get_gint48(tvbuff_t *tvb, const gint offset, const guint encoding);
extern guint64 tvb_get_guint56(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gint64 tvb_get_gint56(tvbuff_t *tvb, const gint offset, const guint encoding);
extern guint64 tvb_get_guint64(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gint64 tvb_get_gint64(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gfloat tvb_get_ieee_float(tvbuff_t *tvb, const gint offset, const guint encoding);
extern gdouble tvb_get_ieee_double(tvbuff_t *tvb, const gint offset, const guint encoding);

/*
 * Fetch 16-bit and 32-bit values in host byte order.
 * Used for some pseudo-headers in pcap/pcapng files, in which the
 * headers are, when capturing, in the byte order of the host, and
 * are converted to the byte order of the host reading the file
 * when reading a capture file.
 * FIXME: not supported
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
#define tvb_get_h_guint16   tvb_get_letohs
#define tvb_get_h_guint32   tvb_get_letohl
#elif G_BYTE_ORDER == G_BIG_ENDIAN
#define tvb_get_h_guint16   tvb_get_ntohs
#define tvb_get_h_guint32   tvb_get_ntohl
#else
#error "Unsupported byte order"
#endif
*/


/* Fetch a time value from an ASCII-style string in the tvb.
 *
 * @param[in] offset The beginning offset in the tvb (cannot be negative)
 * @param[in] length The field's length in the tvb (or -1 for remaining)
 * @param[in] encoding The ENC_* that defines the format (e.g., ENC_ISO_8601_DATE_TIME)
 * @param[in,out] ns The pre-allocated nstime_t that will be set to the decoded value
 * @param[out] endoff if not NULL, should point to a gint that this
 *     routine will then set to be the offset to the character after
 *     the last character used in the conversion. This is useful because
 *     they may not consume the whole section.
 *
 * @return a pointer to the nstime_t passed-in, or NULL on failure; if no
 *    valid conversion could be performed, *endoff is set to 0, and errno will be
 *    EDOM or ERANGE, and the nstime_t* passed-in will be cleared.
 *
 * @note The conversion ignores leading spaces, and will succeed even if it does
 *    not consume the entire string. If you care about such things, always compare
 *    the *endoff to where you expect it to be (namely, offset+length).
 *
 * This routine will not throw an error unless the passed-in arguments are
 * invalid (e.g., offset is beyond the tvb's length).
 *
 * @warning This only works for string encodings which encode ASCII characters in
 * a single byte: ENC_ASCII, ENC_UTF_8, ENC_ISO_8859_*, etc. It does NOT work
 * for purely multi-byte encodings such as ENC_UTF_16, ENC_UCS_*, etc.
 */
extern
nstime_t* tvb_get_string_time(tvbuff_t *tvb, const gint offset, const gint length,
                              const guint encoding, nstime_t* ns, gint *endoff);

/* Similar to above, but returns a GByteArray based on the case-insensitive
 * hex-char strings with optional separators, and with optional leading spaces.
 * The separators allowed are based on the ENC_SEP_* passed in the encoding param.
 *
 * The passed-in bytes is set to the values, and its pointer is also the return
 * value or NULL on error. The GByteArray bytes must be pre-constructed with
 * g_byte_array_new().
 */
extern
GByteArray* tvb_get_string_bytes(tvbuff_t *tvb, const gint offset, const gint length,
                                 const guint encoding, GByteArray* bytes, gint *endoff);

/**
 * Fetch an IPv4 address, in network byte order.
 * We do *not* convert it to host byte order; we leave it in
 * network byte order, as that's what its callers expect. */
extern guint32 tvb_get_ipv4(tvbuff_t *tvb, const gint offset);

/* Fetch an IPv6 address. */
extern void tvb_get_ipv6(tvbuff_t *tvb, const gint offset,
    ws_in6_addr *addr);

/* Fetch a GUID. */
extern void tvb_get_ntohguid(tvbuff_t *tvb, const gint offset,
    e_guid_t *guid);
extern void tvb_get_letohguid(tvbuff_t *tvb, const gint offset,
    e_guid_t *guid);
extern void tvb_get_guid(tvbuff_t *tvb, const gint offset,
    e_guid_t *guid, const guint encoding);

/* Fetch a specified number of bits from bit offset in a tvb.  All of these
 * functions are equivalent, except for the type of the return value.  Note
 * that the parameter encoding (where supplied) is meaningless and ignored */

/* get 1 - 8 bits returned in a guint8 */
extern guint8 tvb_get_bits8(tvbuff_t *tvb, guint bit_offset,
    const gint no_of_bits);
/* get 1 - 16 bits returned in a guint16 */
extern guint16 tvb_get_bits16(tvbuff_t *tvb, guint bit_offset,
    const gint no_of_bits, const guint encoding);
/* get 1 - 32 bits returned in a guint32 */
extern guint32 tvb_get_bits32(tvbuff_t *tvb, guint bit_offset,
    const gint no_of_bits, const guint encoding);
/* get 1 - 64 bits returned in a guint64 */
extern guint64 tvb_get_bits64(tvbuff_t *tvb, guint bit_offset,
    const gint no_of_bits, const guint encoding);

/**
 *  This function has EXACTLY the same behavior as
 *  tvb_get_bits32()
 */
extern guint32 tvb_get_bits(tvbuff_t *tvb, const guint bit_offset,
    const gint no_of_bits, const guint encoding);

/** Returns target for convenience. Does not suffer from possible
 * expense of tvb_get_ptr(), since this routine is smart enough
 * to copy data in chunks if the request range actually exists in
 * different "real" tvbuffs. This function assumes that the target
 * memory is already allocated; it does not allocate or free the
 * target memory. */
extern void *tvb_memcpy(tvbuff_t *tvb, void *target, const gint offset,
    size_t length);

/** Given an allocator scope, a tvbuff, a byte offset, a byte length:
 *
 *    allocate a buffer using the specified scope;
 *
 *    copy the data from the tvbuff specified by the offset and length
 *    into that buffer, using tvb_memcpy();
 *
 *    and return a pointer to the buffer.
 *
 * Throws an exception if the tvbuff ends before the data being copied does.
 *
 * If scope is set to NULL it is the user's responsibility to wmem_free()
 * the memory allocated. Otherwise memory is automatically freed when the
 * scope lifetime is reached.
 */
extern void *tvb_memdup(wmem_allocator_t *scope, tvbuff_t *tvb,
    const gint offset, size_t length);

/** WARNING! This function is possibly expensive, temporarily allocating
 * another copy of the packet data. Furthermore, it's dangerous because once
 * this pointer is given to the user, there's no guarantee that the user will
 * honor the 'length' and not overstep the boundaries of the buffer.
 *
 * If you're thinking of using tvb_get_ptr, STOP WHAT YOU ARE DOING
 * IMMEDIATELY. Go take a break. Consider that tvb_get_ptr hands you
 * a raw, unprotected pointer that you can easily use to create a
 * security vulnerability or otherwise crash Wireshark. Then consider
 * that you can probably find a function elsewhere in this file that
 * does exactly what you want in a much more safe and robust manner.
 *
 * The returned pointer is data that is internal to the tvbuff, so do not
 * attempt to free it. Don't modify the data, either, because another tvbuff
 * that might be using this tvbuff may have already copied that portion of
 * the data (sometimes tvbuff's need to make copies of data, but that's the
 * internal implementation that you need not worry about). Assume that the
 * guint8* points to read-only data that the tvbuff manages.
 *
 * Return a pointer into our buffer if the data asked for via 'offset'/'length'
 * is contiguous (which might not be the case for a "composite" tvbuff). If the
 * data is not contiguous, a tvb_memdup() is called for the entire buffer
 * and the pointer to the newly-contiguous data is returned. This dynamically-
 * allocated memory will be freed when the tvbuff is freed, after the
 * tvbuff_free_cb_t() is called, if any. */
extern const guint8 *tvb_get_ptr(tvbuff_t *tvb, const gint offset,
    const gint length);

/** Find first occurrence of needle in tvbuff, starting at offset. Searches
 * at most maxlength number of bytes; if maxlength is -1, searches to
 * end of tvbuff.
 * Returns the offset of the found needle, or -1 if not found.
 * Will not throw an exception, even if maxlength exceeds boundary of tvbuff;
 * in that case, -1 will be returned if the boundary is reached before
 * finding needle. */
extern gint tvb_find_guint8(tvbuff_t *tvb, const gint offset,
    const gint maxlength, const guint8 needle);

/** Same as tvb_find_guint8() with 16bit needle. */
extern gint tvb_find_guint16(tvbuff_t *tvb, const gint offset,
    const gint maxlength, const guint16 needle);

/** Find first occurrence of any of the needles of the pre-compiled pattern in
 * tvbuff, starting at offset. The passed in pattern must have been "compiled"
 * before-hand, using ws_mempbrk_compile().
 * Searches at most maxlength number of bytes. Returns the offset of the
 * found needle, or -1 if not found and the found needle.
 * Will not throw an exception, even if
 * maxlength exceeds boundary of tvbuff; in that case, -1 will be returned if
 * the boundary is reached before finding needle. */
/* FIXME: May be not supported easily
extern gint tvb_ws_mempbrk_pattern_guint8(tvbuff_t *tvb, const gint offset,
    const gint maxlength, const ws_mempbrk_pattern* pattern, guchar *found_needle);
*/


/** Find size of stringz (NUL-terminated string) by looking for terminating
 * NUL.  The size of the string includes the terminating NUL.
 *
 * If the NUL isn't found, it throws the appropriate exception.
 */
extern guint tvb_strsize(tvbuff_t *tvb, const gint offset);

/** Find size of UCS-2 or UTF-16 stringz (NUL-terminated string) by
 * looking for terminating 16-bit NUL.  The size of the string includes
 * the terminating NUL.
 *
 * If the NUL isn't found, it throws the appropriate exception.
 */
extern guint tvb_unicode_strsize(tvbuff_t *tvb, const gint offset);

/** Find length of string by looking for end of zero terminated string, up to
 * 'maxlength' characters'; if 'maxlength' is -1, searches to end
 * of tvbuff.
 * Returns -1 if 'maxlength' reached before finding EOS. */
extern gint tvb_strnlen(tvbuff_t *tvb, const gint offset,
    const guint maxlength);

/**
 * Format the data in the tvb from offset for size.  Returned string is
 * wmem packet_scoped so call must be in that scope.
 */
extern gchar *tvb_format_text(tvbuff_t *tvb, const gint offset,
    const gint size);

/**
 * Like "tvb_format_text()", but for 'wsp'; don't show
 * the characters as C-style escapes.
 */
extern gchar *tvb_format_text_wsp(wmem_allocator_t* allocator, tvbuff_t *tvb, const gint offset,
    const gint size);

/**
 * Like "tvb_format_text()", but for null-padded strings; don't show
 * the null padding characters as "\000".  Returned string is wmem packet_scoped
 * so call must be in that scope.
 */
/* FIXME : Not supported yet
extern gchar *tvb_format_stringzpad(tvbuff_t *tvb, const gint offset,
    const gint size);
*/
/**
 * Like "tvb_format_text_wsp()", but for null-padded strings; don't show
 * the null padding characters as "\000".
 */
/* FIXME:
extern gchar *tvb_format_stringzpad_wsp(wmem_allocator_t* allocator, tvbuff_t *tvb, const gint offset,
    const gint size);
*/
/**
 * Given an allocator scope, a tvbuff, a byte offset, a byte length, and
 * a string encoding, with the specified offset and length referring to
 * a string in the specified encoding:
 *
 *    allocate a buffer using the specified scope;
 *
 *    convert the string from the specified encoding to UTF-8, possibly
 *    mapping some characters or invalid octet sequences to the Unicode
 *    REPLACEMENT CHARACTER, and put the resulting UTF-8 string, plus a
 *    trailing '\0', into that buffer;
 *
 *    and return a pointer to the buffer.
 *
 * Throws an exception if the tvbuff ends before the string does.
 *
 * If scope is set to NULL it is the user's responsibility to wmem_free()
 * the memory allocated. Otherwise memory is automatically freed when the
 * scope lifetime is reached.
 */
extern guint8 *tvb_get_string_enc(wmem_allocator_t *scope,
    tvbuff_t *tvb, const gint offset, const gint length, const guint encoding);

/**
 * Given an allocator scope, a tvbuff, a bit offset, and a length in
 * 7-bit characters (not octets!), with the specified offset and
 * length referring to a string in the 3GPP TS 23.038 7bits encoding:
 *
 *    allocate a buffer using the specified scope;
 *
 *    convert the string from the specified encoding to UTF-8, possibly
 *    mapping some characters or invalid octet sequences to the Unicode
 *    REPLACEMENT CHARACTER, and put the resulting UTF-8 string, plus a
 *    trailing '\0', into that buffer;
 *
 *    and return a pointer to the buffer.
 *
 * Throws an exception if the tvbuff ends before the string does.
 *
 * If scope is set to NULL it is the user's responsibility to wmem_free()
 * the memory allocated. Otherwise memory is automatically freed when the
 * scope lifetime is reached.
 */
extern gchar *tvb_get_ts_23_038_7bits_string(wmem_allocator_t *scope,
    tvbuff_t *tvb, const gint bit_offset, gint no_of_chars);

/**
 * Given an allocator scope, a tvbuff, a bit offset, and a length in
 * 7-bit characters (not octets!), with the specified offset and
 * length referring to a string in the ASCII 7bits encoding:
 *
 *    allocate a buffer using the specified scope;
 *
 *    convert the string from the specified encoding to UTF-8, possibly
 *    mapping some characters or invalid octet sequences to the Unicode
 *    REPLACEMENT CHARACTER, and put the resulting UTF-8 string, plus a
 *    trailing '\0', into that buffer;
 *
 *    and return a pointer to the buffer.
 *
 * Throws an exception if the tvbuff ends before the string does.
 *
 * If scope is set to NULL it is the user's responsibility to wmem_free()
 * the memory allocated. Otherwise memory is automatically freed when the
 * scope lifetime is reached.
 */
extern gchar *tvb_get_ascii_7bits_string(wmem_allocator_t *scope,
    tvbuff_t *tvb, const gint bit_offset, gint no_of_chars);

/**
 * Given an allocator scope, a tvbuff, a byte offset, a byte length, and
 * a string encoding, with the specified offset and length referring to
 * a null-padded string in the specified encoding:
 *
 *    allocate a buffer using the specified scope;
 *
 *    convert the string from the specified encoding to UTF-8, possibly
 *    mapping some characters or invalid octet sequences to the Unicode
 *    REPLACEMENT CHARACTER, and put the resulting UTF-8 string, plus a
 *    trailing '\0', into that buffer;
 *
 *    and return a pointer to the buffer.
 *
 * Throws an exception if the tvbuff ends before the string does.
 *
 * If scope is set to NULL it is the user's responsibility to wmem_free()
 * the memory allocated. Otherwise memory is automatically freed when the
 * scope lifetime is reached.
 */
extern guint8 *tvb_get_stringzpad(wmem_allocator_t *scope,
    tvbuff_t *tvb, const gint offset, const gint length, const guint encoding);

/**
 * Given an allocator scope, a tvbuff, a byte offset, a pointer to a
 * gint, and a string encoding, with the specified offset referring to
 * a null-terminated string in the specified encoding:
 *
 *    find the length of that string (and throw an exception if the tvbuff
 *    ends before we find the null);
 *
 *    allocate a buffer using the specified scope;
 *
 *    convert the string from the specified encoding to UTF-8, possibly
 *    mapping some characters or invalid octet sequences to the Unicode
 *    REPLACEMENT CHARACTER, and put the resulting UTF-8 string, plus a
 *    trailing '\0', into that buffer;
 *
 *    if the pointer to the gint is non-null, set the gint to which it
 *    points to the length of the string;
 *
 *    and return a pointer to the buffer.
 *
 * Throws an exception if the tvbuff ends before the string does.
 *
 * If scope is set to NULL it is the user's responsibility to wmem_free()
 * the memory allocated. Otherwise memory is automatically freed when the
 * scope lifetime is reached.
 */
extern guint8 *tvb_get_stringz_enc(wmem_allocator_t *scope,
    tvbuff_t *tvb, const gint offset, gint *lengthp, const guint encoding);

/**
 * Given a tvbuff and an offset, with the offset assumed to refer to
 * a null-terminated string, find the length of that string (and throw
 * an exception if the tvbuff ends before we find the null), allocate
 * a buffer big enough to hold the string, copy the string into it,
 * and return a pointer to the string.  Also return the length of the
 * string (including the terminating null) through a pointer.
 *
 * This returns a constant (unmodifiable) string that does not need
 * to be freed; instead, it will automatically be freed once the next
 * packet is dissected.
 *
 * It is slightly more efficient than the other routines, but does *NOT*
 * do any translation to UTF-8 - the string consists of the raw octets
 * of the string, in whatever encoding they happen to be in, and, if
 * the string is not valid in that encoding, with invalid octet sequences
 * as they are in the packet.
 */
extern const guint8 *tvb_get_const_stringz(tvbuff_t *tvb,
    const gint offset, gint *lengthp);

/** Looks for a stringz (NUL-terminated string) in tvbuff and copies
 * no more than bufsize number of bytes, including terminating NUL, to buffer.
 * Returns length of string (not including terminating NUL), or -1 if the
 * string was truncated in the buffer due to not having reached the terminating
 * NUL.  In this way, it acts like g_snprintf().
 *
 * When processing a packet where the remaining number of bytes is less
 * than bufsize, an exception is not thrown if the end of the packet
 * is reached before the NUL is found. If no NUL is found before reaching
 * the end of the short packet, -1 is still returned, and the string
 * is truncated with a NUL, albeit not at buffer[bufsize - 1], but
 * at the correct spot, terminating the string.
 */
extern gint tvb_get_nstringz(tvbuff_t *tvb, const gint offset,
    const guint bufsize, guint8 *buffer);

/** Like tvb_get_nstringz(), but never returns -1. The string is guaranteed to
 * have a terminating NUL. If the string was truncated when copied into buffer,
 * a NUL is placed at the end of buffer to terminate it.
 *
 * bufsize MUST be greater than 0.
 */
extern gint tvb_get_nstringz0(tvbuff_t *tvb, const gint offset,
    const guint bufsize, guint8 *buffer);

/*
 * Given a tvbuff, an offset into the tvbuff, a buffer, and a buffer size,
 * extract as many raw bytes from the tvbuff, starting at the offset,
 * as 1) are available in the tvbuff and 2) will fit in the buffer, leaving
 * room for a terminating NUL.
 */
extern gint tvb_get_raw_bytes_as_string(tvbuff_t *tvb, const gint offset, char *buffer, size_t bufsize);

/** Iterates over the provided portion of the tvb checking that each byte
* is an ascii printable character.
* Returns TRUE if all bytes are printable, FALSE otherwise
*/
extern gboolean tvb_ascii_isprint(tvbuff_t *tvb, const gint offset,
	const gint length);

/**
 * Given a tvbuff, an offset into the tvbuff, and a length that starts
 * at that offset (which may be -1 for "all the way to the end of the
 * tvbuff"), find the end of the (putative) line that starts at the
 * specified offset in the tvbuff, going no further than the specified
 * length.
 *
 * Return the length of the line (not counting the line terminator at
 * the end), or, if we don't find a line terminator:
 *
 *  if "deseg" is true, return -1;
 *
 *  if "deseg" is false, return the amount of data remaining in
 *  the buffer.
 *
 * Set "*next_offset" to the offset of the character past the line
 * terminator, or past the end of the buffer if we don't find a line
 * terminator.  (It's not set if we return -1.)
 */
extern gint tvb_find_line_end(tvbuff_t *tvb, const gint offset, int len,
    gint *next_offset, const gboolean desegment);

/**
 * Given a tvbuff, an offset into the tvbuff, and a length that starts
 * at that offset (which may be -1 for "all the way to the end of the
 * tvbuff"), find the end of the (putative) line that starts at the
 * specified offset in the tvbuff, going no further than the specified
 * length.
 *
 * However, treat quoted strings inside the buffer specially - don't
 * treat newlines in quoted strings as line terminators.
 *
 * Return the length of the line (not counting the line terminator at
 * the end), or the amount of data remaining in the buffer if we don't
 * find a line terminator.
 *
 * Set "*next_offset" to the offset of the character past the line
 * terminator, or past the end of the buffer if we don't find a line
 * terminator.
 */
extern gint tvb_find_line_end_unquoted(tvbuff_t *tvb, const gint offset,
    int len, gint *next_offset);

/**
 * Copied from the mgcp dissector. (This function should be moved to /epan )
 * tvb_skip_wsp - Returns the position in tvb of the first non-whitespace
 *                character following offset or offset + maxlength -1 whichever
 *                is smaller.
 *
 * Parameters:
 * tvb - The tvbuff in which we are skipping whitespace.
 * offset - The offset in tvb from which we begin trying to skip whitespace.
 * maxlength - The maximum distance from offset that we may try to skip
 * whitespace.
 *
 * Returns: The position in tvb of the first non-whitespace
 *          character following offset or offset + maxlength -1 whichever
 *          is smaller.
 */

extern gint tvb_skip_wsp(tvbuff_t *tvb, const gint offset,
    const gint maxlength);

extern gint tvb_skip_wsp_return(tvbuff_t *tvb, const gint offset);

// int tvb_skip_guint8(tvbuff_t *tvb, int offset, const int maxlength, const guint8 ch);

/**
* Given a tvbuff, an offset into the tvbuff, and a length that starts
* at that offset (which may be -1 for "all the way to the end of the
* tvbuff"), find the end of the token that starts at the
* specified offset in the tvbuff, going no further than the specified
* length.
*
* Return the length of the token, or, if we don't find a terminator:
*
*  if "deseg" is true, return -1;
*
*  if "deseg" is false, return the amount of data remaining in
*  the buffer.
*
* Set "*next_offset" to the offset of the character past the
* terminator, or past the end of the buffer if we don't find a line
* terminator.  (It's not set if we return -1.)
*/
extern int tvb_get_token_len(tvbuff_t *tvb, const gint offset, int len, gint *next_offset, const gboolean desegment);

/**
 * Call strncmp after checking if enough chars left, returning 0 if
 * it returns 0 (meaning "equal") and -1 otherwise, otherwise return -1.
 */
extern gint tvb_strneql(tvbuff_t *tvb, const gint offset,
    const gchar *str, const size_t size);

/**
 * Call g_ascii_strncasecmp after checking if enough chars left, returning
 * 0 if it returns 0 (meaning "equal") and -1 otherwise, otherwise return -1.
 */
extern gint tvb_strncaseeql(tvbuff_t *tvb, const gint offset,
    const gchar *str, const size_t size);

/**
 * Call memcmp after checking if enough chars left, returning 0 if
 * it returns 0 (meaning "equal") and -1 otherwise, otherwise return -1.
 */
extern gint tvb_memeql(tvbuff_t *tvb, const gint offset,
    const guint8 *str, size_t size);

/**
 * Format a bunch of data from a tvbuff as bytes, returning a pointer
 * to the string with the formatted data, with "punct" as a byte
 * separator.
 */
extern gchar *tvb_bytes_to_str_punct(wmem_allocator_t *scope, tvbuff_t *tvb, const gint offset,
    const gint len, const gchar punct);

/**
 * Format a bunch of data from a tvbuff as bytes, returning a pointer
 * to the string with the formatted data.
 */
extern gchar *tvb_bytes_to_str(wmem_allocator_t *allocator, tvbuff_t *tvb,
    const gint offset, const gint len);


extern const gchar *tvb_bcd_dig_to_wmem_packet_str(tvbuff_t *tvb,
    const gint offset, const gint len, const dgt_set_t *dgt,
    gboolean skip_first);

/** Locate a sub-tvbuff within another tvbuff, starting at position
 * 'haystack_offset'. Returns the index of the beginning of 'needle' within
 * 'haystack', or -1 if 'needle' is not found. The index is relative
 * to the start of 'haystack', not 'haystack_offset'. */
extern gint tvb_find_tvb(tvbuff_t *haystack_tvb, tvbuff_t *needle_tvb,
    const gint haystack_offset);

/* From tvbuff_zlib.c */

/**
 * Uncompresses a zlib compressed packet inside a tvbuff at offset with
 * length comprlen.  Returns an uncompressed tvbuffer if uncompression
 * succeeded or NULL if uncompression failed.
 */
extern tvbuff_t *tvb_uncompress(tvbuff_t *tvb, const int offset,
    int comprlen);

/**
 * Uncompresses a zlib compressed packet inside a tvbuff at offset with
 * length comprlen.  Returns an uncompressed tvbuffer attached to tvb if
 * uncompression succeeded or NULL if uncompression failed.
 */
extern tvbuff_t *tvb_child_uncompress(tvbuff_t *parent, tvbuff_t *tvb,
    const int offset, int comprlen);

/* From tvbuff_brotli.c */

/**
 * Uncompresses a brotli compressed packet inside a tvbuff at offset with
 * length comprlen.  Returns an uncompressed tvbuffer if uncompression
 * succeeded or NULL if uncompression failed.
 */
extern tvbuff_t *tvb_uncompress_brotli(tvbuff_t *tvb, const int offset,
    int comprlen);

/**
 * Uncompresses a brotli compressed packet inside a tvbuff at offset with
 * length comprlen.  Returns an uncompressed tvbuffer attached to tvb if
 * uncompression succeeded or NULL if uncompression failed.
 */
extern tvbuff_t *tvb_child_uncompress_brotli(tvbuff_t *parent, tvbuff_t *tvb,
    const int offset, int comprlen);

/* From tvbuff_lz77.c */

/**
 * Uncompresses a Microsoft Plain LZ77 compressed payload inside a
 * tvbuff at offset with length comprlen.  Returns an uncompressed
 * tvbuffer if uncompression succeeded or NULL if uncompression
 * failed.
 */
extern tvbuff_t *tvb_uncompress_lz77(tvbuff_t *tvb,
    const int offset, int comprlen);

/**
 * Uncompresses a Microsoft Plain LZ77 compressed payload inside a
 * tvbuff at offset with length comprlen.  Returns an uncompressed
 * tvbuffer attached to tvb if uncompression succeeded or NULL if
 * uncompression failed.
 */
extern tvbuff_t *tvb_child_uncompress_lz77(tvbuff_t *parent,
     tvbuff_t *tvb, const int offset, int comprlen);

/* From tvbuff_lz77huff.c */

/**
 * Uncompresses a Microsoft LZ77+Huffman compressed payload inside a
 * tvbuff at offset with length comprlen.  Returns an uncompressed
 * tvbuffer if uncompression succeeded or NULL if uncompression
 * failed.
 */
extern tvbuff_t *tvb_uncompress_lz77huff(tvbuff_t *tvb,
    const int offset, int comprlen);

/**
 * Uncompresses a Microsoft LZ77+Huffman compressed payload inside a
 * tvbuff at offset with length comprlen.  Returns an uncompressed
 * tvbuffer attached to tvb if uncompression succeeded or NULL if
 * uncompression failed.
 */
extern tvbuff_t *tvb_child_uncompress_lz77huff(tvbuff_t *parent,
    tvbuff_t *tvb, const int offset, int comprlen);

/* From tvbuff_lznt1.c */

/**
 * Uncompresses a Microsoft LZNT1 compressed payload inside
 * a tvbuff at offset with length comprlen.  Returns an uncompressed
 * tvbuffer if uncompression succeeded or NULL if uncompression
 * failed.
 */
extern tvbuff_t *tvb_uncompress_lznt1(tvbuff_t *tvb,
    const int offset, int comprlen);

/**
 * Uncompresses a Microsoft LZNT1 compressed payload inside
 * a tvbuff at offset with length comprlen.  Returns an uncompressed
 * tvbuffer if uncompression succeeded or NULL if uncompression
 * failed.
 */
extern tvbuff_t *tvb_child_uncompress_lznt1(tvbuff_t *parent,
    tvbuff_t *tvb, const int offset, int comprlen);

/* From tvbuff_base64.c */

/** Return a tvb that contains the binary representation of a base64
 *  string as a child of the indicated tvb.
 *
 * @param parent The parent tvbuff.
 * @param base64 The base64 encoded string which binary representation will be
 *               returned in the child tvb.
 *
 * @return   A tvb with the binary representation of the base64 decoded string.
 */
// FIXME : extern tvbuff_t* base64_to_tvb(tvbuff_t *parent, const char *base64);

/**
 * Extract a variable length integer from a tvbuff.
 * Each byte in a varint, except the last byte, has the most significant bit (msb)
 * set -- this indicates that there are further bytes to come. For example,
 *   1010 1100 0000 0010 is 300
 *
 * @param tvb The tvbuff in which we are extracting integer.
 * @param offset The offset in tvb from which we begin trying to extract integer.
 * @param maxlen The maximum distance from offset that we may try to extract integer
 * @param value  if parsing succeeds, parsed varint will store here.
 * @param encoding The ENC_* that defines the format (e.g., ENC_VARINT_PROTOBUF, ENC_VARINT_QUIC, ENC_VARINT_ZIGZAG)
 * @return   the length of this varint in tvb. 0 means parsing failed.
 */
extern guint tvb_get_varint(tvbuff_t *tvb, guint offset, guint maxlen, guint64 *value, const guint encoding);

/************** END OF ACCESSORS ****************/

/** @} */


/*
 * Editor modelines  -  https://www.wireshark.org/tools/modelines.html
 *
 * Local variables:
 * c-basic-offset: 4
 * tab-width: 8
 * indent-tabs-mode: nil
 * End:
 *
 * vi: set shiftwidth=4 tabstop=8 expandtab:
 * :indentSize=4:tabSize=8:noTabs=true:
 */
"""



import os
import sys
import time
import struct
import socket
from datetime import datetime as dt
import warnings

_MAX_TO_PROCESS = 10000000

from wishpy.wireshark.lib.dissector import WishpyDissectorFile
from wishpy.wireshark.lib.dissector import setup_process, cleanup_process

if __name__ == '__main__':

    if not len(sys.argv) >= 2:
        print("Usage: tshark.py <filepath>")
        sys.exit(1)

    input_filepath = sys.argv[1]

    setup_process()

    dissector = WishpyDissectorFile(input_filepath)

    then = dt.now()

    try:
        for dissected in dissector.run(count=0, skip=0):
            print(dissected)

    except KeyboardInterrupt:
        cleanup_process()

    now = dt.now()




wsutil_ws_mempbrk_h_types_cdef = """

/* FIXME: Later may be
#ifdef HAVE_SSE4_2
#include <emmintrin.h>
#endif
*/

/** The pattern object used for ws_mempbrk_exec().
 */
typedef long long __m128i;

typedef struct {
    gchar patt[256];
/*  FIXME : Later may be
#ifdef HAVE_SSE4_2
#endif
*/
    gboolean use_sse42;
    __m128i mask;
} ws_mempbrk_pattern;


"""


wsutil_ws_mempbrk_h_funcs_cdef = """

/** Compile the pattern for the needles to find using ws_mempbrk_exec().
 */
extern void ws_mempbrk_compile(ws_mempbrk_pattern* pattern, const gchar *needles);

/** Scan for the needles specified by the compiled pattern.
 */
extern const guint8 *ws_mempbrk_exec(const guint8* haystack, size_t haystacklen, const ws_mempbrk_pattern* pattern, guchar *found_needle);

"""



wsutil_inet_ipv6_h_cdef = """

#define IPv6_ADDR_SIZE  16

#define IPv6_HDR_SIZE           40
#define IPv6_FRAGMENT_HDR_SIZE  8

typedef struct e_in6_addr {
    guint8 bytes[16];           /* 128 bit IPv6 address */
} ws_in6_addr;

/*
 * Definition for internet protocol version 6.
 * RFC 2460
 */
struct ws_ip6_hdr {
    guint32     ip6h_vc_flow;           /* version, class, flow */
    guint16     ip6h_plen;              /* payload length */
    guint8      ip6h_nxt;               /* next header */
    guint8      ip6h_hlim;              /* hop limit */
    ws_in6_addr ip6h_src;               /* source address */
    ws_in6_addr ip6h_dst;               /* destination address */
};

/*
 * Extension Headers
 */

struct ip6_ext {
    guchar ip6e_nxt;
    guchar ip6e_len;
};

/* Routing header */
struct ip6_rthdr {
    guint8 ip6r_nxt;        /* next header */
    guint8 ip6r_len;        /* length in units of 8 octets */
    guint8 ip6r_type;       /* routing type */
    guint8 ip6r_segleft;    /* segments left */
    /* followed by routing type specific data */
};

/* Type 0 Routing header */
struct ip6_rthdr0 {
    guint8 ip6r0_nxt;       /* next header */
    guint8 ip6r0_len;       /* length in units of 8 octets */
    guint8 ip6r0_type;      /* always zero */
    guint8 ip6r0_segleft;   /* segments left */
    guint8 ip6r0_reserved;  /* reserved field */
    guint8 ip6r0_slmap[3];  /* strict/loose bit map */
    /* followed by up to 127 addresses */
    ws_in6_addr ip6r0_addr[1];
};

/* Fragment header */
struct ip6_frag {
    guint8  ip6f_nxt;       /* next header */
    guint8  ip6f_reserved;  /* reserved field */
    guint16 ip6f_offlg;     /* offset, reserved, and flag */
    guint32 ip6f_ident;     /* identification */
};

#define IP6F_OFF_MASK           0xfff8 /* mask out offset from _offlg */
#define IP6F_RESERVED_MASK      0x0006 /* reserved bits in ip6f_offlg */
#define IP6F_MORE_FRAG          0x0001 /* more-fragments flag */


"""

# FIXME: Following should go in set_source may be?

wsutil_inet_ipv6_source = """
/**
 * Unicast Scope
 * Note that we must check topmost 10 bits only, not 16 bits (see RFC2373).
 */
gboolean in6_addr_is_linklocal(const ws_in6_addr *a)
{
    return (a->bytes[0] == 0xfe) && ((a->bytes[1] & 0xc0) == 0x80);
}

gboolean in6_addr_is_sitelocal(const ws_in6_addr *a)
{
    return (a->bytes[0] == 0xfe) && ((a->bytes[1] & 0xc0) == 0xc0);
}

/**
 * Multicast
 */
gboolean in6_addr_is_multicast(const ws_in6_addr *a)
{
    return a->bytes[0] == 0xff;
}
"""



"""
Our Dissector API.
"""
import socket
import struct
import json
import time
from datetime import datetime as dt

from ._wrapper import *


class WishpyEpanLibUninitializedError(Exception):
    pass

class WishpyEpanLibAlreadyInitialized(Exception):
    pass

class WishpyErrorWthOpen(Exception):
    pass

_EPAN_LIB_INITIALIZED = False

class WishpyDissectorBase:
    """ A Class that wraps the underlying dissector from epan module of
    `libwireshark`. Right now this simply prints the dissector tree.
    """

    # Below are some dict's required for printing few packet types
    hfbases = {
            epan_lib.BASE_NONE : ('{:d}', False), # Not sure how this is to be treated
            epan_lib.BASE_DEC : ('{:d}', False),
            epan_lib.BASE_HEX : ('0x{:x}', True),
            epan_lib.BASE_OCT : ('{:o}', True),
            epan_lib.BASE_DEC_HEX: ('{:d}(0x{:x})', True),
            epan_lib.BASE_HEX_DEC: ('0x{:x}({:d})', True),
            epan_lib.BASE_PT_TCP: ('{:d}', False),
            epan_lib.BASE_PT_UDP: ('{:d}', False),
            epan_lib.BASE_PT_SCTP: ('{:d}', False)

            }

    @classmethod
    def func_not_supported(cls, *args):
        return "Not Supported"

    @classmethod
    def epan_ether_to_str(cls, fvalue, ftype, display):
        """ Converting Ethernet addresses to String"""

        eth_bytes = fvalue.value.bytes

        display_bytes = []
        for i in range(eth_bytes.len):
            display_byte = "{:02X}".format(eth_bytes.data[i])
            display_bytes.append(display_byte)

        return ":".join(display_bytes), True

    epan_bytes_to_str = epan_ether_to_str

    @classmethod
    def epan_str_to_str(cls, fvalue, ftype, display):
        """Converting epan 'char *' to Python String"""

        value = fvalue.value.string

        # If display is BASE_NONE, utf-8 is fine, but for others, not always! (GSM-Unicode)
        if display == 0:
            x = epan_ffi.string(value).decode("utf-8").\
                    replace('\\', '\\\\').\
                    replace('"', '\\"')
            return repr(x), True
        else:
            try:
                x = epan_ffi.string(value).decode("utf-8").\
                        replace('\\', '\\\\').\
                        replace('"', '\\"')
                return repr(x), True
            except:
                return "Cannot Decode"

    @classmethod
    def epan_ipv4_to_str(cls, fvalue, ftype, display):
        """Converting IP Address to Python String"""

        ipv4 = fvalue.value.ipv4

        return socket.inet_ntoa(struct.pack('!I', ipv4.addr)), True

    @classmethod
    def epan_bool_to_str(cls, fvalue, ftype, display, on_off=False, json_compat=True):
        """ Converting `gboolean` to String"""

        if on_off and json_compat:
            raise ValueError("Specify Either on_off or json_compat not both.")

        value = bool(fvalue.value.uinteger)

        if value:
            if json_compat:
                return "true", False
            if on_off:
                return "ON", True
        else:
            if json_compat:
                return "false", False
            if on_off:
                return "OFF", True

        return "{}".format(value)

    @classmethod
    def epan_int_to_str(cls, fvalue, ftype, display):
        """Converting Integer to String, using BASE_* property."""

        try:
            # We are not displaying 'Extended string for the value
            if display & epan_lib.BASE_EXT_STRING:
                display ^= epan_lib.BASE_EXT_STRING

            base_format, quote = cls.hfbases[display]
        except:
            return "type: {} display: {} Not Supported".format(ftype, display), True

        return base_format.format(fvalue.value.uinteger,
                fvalue.value.uinteger), quote

    @classmethod
    def epan_abstime_to_str(cls, fvalue, ftype, display):
        timeval = fvalue.value.time
        timeval = timeval.secs + timeval.nsecs / 1000000000
        value = dt.strftime(dt.fromtimestamp(timeval), '%d-%b-%Y %H:%M:%S.%f %Z')
        return value, False


    @classmethod
    def epan_reltime_to_str(cls, fvalue, ftype, display):
        value = fvalue.value.time
        return "{:.9f}".format(value.secs + value.nsecs / 1000000000), False

    @classmethod
    def value_to_str(cls, finfo):
        """
        Returns string representation of the `finfo.value`
        """

        fvalue = finfo.value
        ftype = finfo.hfinfo[0].type
        display = finfo.hfinfo[0].display
        abbrev = epan_ffi.string(finfo.hfinfo[0].abbrev).decode()

        epan_int_types = [
                epan_lib.FT_INT8,
                epan_lib.FT_INT16,
                epan_lib.FT_INT32,
                epan_lib.FT_INT40,
                epan_lib.FT_INT48,
                epan_lib.FT_INT56,
                epan_lib.FT_INT64]

        epan_uint32_types = [
                epan_lib.FT_CHAR,
                epan_lib.FT_UINT8,
                epan_lib.FT_UINT16,
                epan_lib.FT_UINT32,
                epan_lib.FT_FRAMENUM]

        epan_uint_types = epan_uint32_types + \
                [ epan_lib.FT_UINT40, epan_lib.FT_UINT48,
                epan_lib.FT_UINT56, epan_lib.FT_UINT64]

        epan_all_int_types = epan_int_types + epan_uint_types

        if ftype in epan_all_int_types:
            return cls.epan_int_to_str(fvalue, ftype, display)

        if ftype == epan_lib.FT_ETHER:
            return cls.epan_ether_to_str(fvalue, ftype, display)

        if ftype == epan_lib.FT_IPv4:
            return cls.epan_ipv4_to_str(fvalue, ftype, display)

        if ftype == epan_lib.FT_BOOLEAN:
            return cls.epan_bool_to_str(fvalue, ftype, display)

        if ftype == epan_lib.FT_STRING:
            return cls.epan_str_to_str(fvalue, ftype, display)

        if ftype == epan_lib.FT_BYTES:
            return cls.epan_bytes_to_str(fvalue, ftype, display)

        if ftype == epan_lib.FT_RELATIVE_TIME:
            return cls.epan_reltime_to_str(fvalue, ftype, display)

        if ftype == epan_lib.FT_ABSOLUTE_TIME:
            return cls.epan_abstime_to_str(fvalue, ftype, display)

        if ftype in [epan_lib.FT_NONE, epan_lib.FT_PROTOCOL]:
            return None, False

        return "{} {}".format(ftype, display), True

    @classmethod
    def print_dissected_tree(cls, node_ptr, data_ptr):
        """
        Returns a string that represents a dissected tree."""

        return_str = ""

        if data_ptr == epan_ffi.NULL:
            level = 1
        else:
            level = epan_ffi.cast('int *', data_ptr)[0]

        return_str = ""
        node = node_ptr[0]
        finfo = node.finfo
        if finfo != epan_ffi.NULL:

            hfinfo = finfo.hfinfo[0]
            abbrev = epan_ffi.string(hfinfo.abbrev).decode()
            abbrev_str = "\"{!s}\"".format(abbrev)
            return_str += abbrev_str + ": "
            finfo_display_str, quote = cls.value_to_str(finfo)
            if finfo_display_str:
                if quote:
                    finfo_display_str = "\"{!s}\"".format(finfo_display_str)
        else:
            finfo_display_str = ""

        if finfo_display_str is not None:
            return_str += finfo_display_str

        data_ptr_new = epan_ffi.new('int *')
        data_ptr_new[0] = level + 1
        child = node.first_child
        if child != epan_ffi.NULL:

            if finfo_display_str:
                return_str += ",\n"

                abbrev_tree = abbrev + "_tree"
                abbrev_tree_str = "\"{!s}\"".format(abbrev_tree)
                return_str += "  " * (level -1)
                return_str += abbrev_tree_str + " : "

            return_str += "{ "
            return_str += "\n"
            while child != epan_ffi.NULL:
                return_str += "  " * level
                return_str += cls.print_dissected_tree(child, data_ptr_new)
                child = child.next
            return_str += "  " * level

            return_str += "\n"
            return_str += "  " * (level-1)
            return_str += "}"
        else: # child is not None So we have someone who's FT_NONE, FT_PROTOCOL and no tree?
            if not finfo_display_str:
                return_str += "\"\""
        if node.next != epan_ffi.NULL:
            return_str += ",\n"

        return return_str

    @classmethod
    def packet_to_json(cls, handle_ptr):
        """ An example method that depicts how to use internal dissector API."""

        dissector = handle_ptr[0]

        # FIXME: following should be like json dumps
        s = cls.print_dissected_tree(dissector.tree, epan_ffi.NULL)
        try:
            x = json.loads(s, strict=False)
        except Exception as e:
            print(s)
            return ""

        return s

    def run(self, *args, **kw):
        """A generator function `yield`ing at-least the dissected packets.

        Implementing this as a generator function helps one to run code
        that looks like

        >>> for dissected in dissector.run(count=1):
            # do stuff with the dissected packet

        This is particularly convenient while performing live capture on
        an interface or dissecting a huge file.
        """
        raise NotImplemented("Derived Classes need to implement this.")

class WishpyDissectorFile(WishpyDissectorBase):
    """Dissector class for PCAP Files.
    """

    def __init__(self, filename):
        self.__filename = filename

    def run(self, count=0, skip=-1):
        """
        Actual function that performs the Dissection. Right now since we are
        only supporting dissecting packets from Wiretap supported files,
        only dissects packets from a pcap(ish) file.
        """

        if not _EPAN_LIB_INITIALIZED:
            raise WishpyEpanLibUninitializedError(
                    "Epan Library Not initialized. Did you call setup_process()"
                    )
        # FIXME: dissector.run can be run only once right now
        # FIXME: Pass errno / errstr ourselves to get the error to be passed
        # to the Exception handler

        # FIXME: Do this as a context manager
        wth, wth_filetype = wtap_open_file_offline(self.__filename)
        if wth is None:
            raise WishpyErrorWthOpen()

        try:
            yield from epan_perform_dissection(wth, wth_filetype,
                    self.packet_to_json, count, skip)
        except Exception as e:
            print(e)
            pass
        finally:
            # If we don't close `wtap` here, outer `cleanup_process` croaks
            wtap_close(wth)

        return

class WishpyDissectorQueue(WishpyDissectorBase):
    """Dissector class for packets received from a Queue(ish) object.
    """

    def dissect_one_packet(self):
        """Dissects a single packet

        This should typically call `fetch` and the perform dissection. All
        the queue based `dissectors` will dissect one packet at a time, so
        it's better that this function is in the base class.
        """
        hdr, data = self.fetch()
        d = epan_perform_one_packet_dissection(self._packets_fetched,
                hdr, data, self.packet_to_json)

        return hdr, data, d

    def fetch(self):
        """Implement this function to fetch a single packet from the queue.

        Implementation of this function should return the object of the type
        `Hdr` and `PacketData`
        """
        raise NotImplemented("Derived Classes Need to implement this.")


class WishpyDissectorQueuePython(WishpyDissectorQueue):

    def __init__(self, queue):
        self.__queue = queue
        self.__running = False
        self.__stop_requested = False
        self._packets_fetched = 0

    def fetch(self):
        """Blocking Fetch from a Python Queue.
        """
        hdr, data = self.__queue.get()
        self._packets_fetched += 1
        return hdr, data

    def __iter__(self):
        return self

    def __next__(self):
        """Returns next `fetch`ed packet. (Blocking)
        """
        return self.dissect_one_packet()

    def run(self, count=0):
        """yield's the packet, up to maximum of `count` packets.

        if count is <= 0, infinite iterator.
        """

        self.__running = True

        fetched = 0
        while True:
            fetched += 1
            hdr, data, d = self.dissect_one_packet()

            x = yield (hdr, data, d)

            if self.__stop_requested == True:
                break

            if x and x.lower() == 'stop':
                break

            if fetched == count:
                break

        self.__running = False

    def stop(self):
        """Stop's the generator by setting internal state."""
        self.__stop_requested = True


def setup_process():
    """
    This method should be called once per process (note: Not thread.) This
    will perform underlying library initialization, so that eventually
    dissectors can `run`.
    """

    global _EPAN_LIB_INITIALIZED

    if _EPAN_LIB_INITIALIZED:
        raise WishpyEpanLibAlreadyInitialized(
                "Epan Library already initialized. setup_process() "\
                        "should be called only once per process.")

    _EPAN_LIB_INITIALIZED = perform_epan_wtap_init()


def cleanup_process():
    """
    Per process cleanup. de-init of epan/wtap modules.
    """

    perform_epan_wtap_cleanup()

    global _EPAN_LIB_INITIALIZED
    _EPAN_LIB_INITIALIZED = False


__all__ = ['WishpyDissectorFile', 'WishpyDissectorQueue',
        'WishpyDissectorQueuePython']



epan_prefs_h_cdef = """

/* prefs.h
 * Definitions for preference handling routines
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#define PR_DEST_CMD  0
#define PR_DEST_FILE 1

#define DEF_WIDTH 750
#define DEF_HEIGHT 550

#define MAX_VAL_LEN  1024

#define RTP_PLAYER_DEFAULT_VISIBLE 4
#define TAP_UPDATE_DEFAULT_INTERVAL 3000
#define ST_DEF_BURSTRES 5
#define ST_DEF_BURSTLEN 100
#define ST_MAX_BURSTRES 600000 /* somewhat arbirary limit of 10 minutes */
#define ST_MAX_BURSTBUCKETS 100 /* somewhat arbirary limit - more buckets degrade performance */

struct epan_uat;
struct _e_addr_resolve;

/*
 * Convert a string listing name resolution types to a bitmask of
 * those types.
 *
 * Set "*name_resolve" to the bitmask, and return '\0', on success;
 * return the bad character in the string on error.
 */
extern
char string_to_name_resolve(const char *string, struct _e_addr_resolve *name_resolve);

/*
 * Modes for the starting directory in File Open dialogs.
 */
#define FO_STYLE_LAST_OPENED    0 /* start in last directory we looked at */
#define FO_STYLE_SPECIFIED      1 /* start in specified directory */

/*
 * Toolbar styles.
 */
#define TB_STYLE_ICONS          0
#define TB_STYLE_TEXT           1
#define TB_STYLE_BOTH           2

/*
 * Types of layout of summary/details/hex panes.
 */
typedef enum {
    layout_unused,  /* entry currently unused */
    layout_type_5,
    layout_type_2,
    layout_type_1,
    layout_type_4,
    layout_type_3,
    layout_type_6,
    layout_type_max
} layout_type_e;

/*
 * Types of pane.
 */
typedef enum {
    layout_pane_content_none,
    layout_pane_content_plist,
    layout_pane_content_pdetails,
    layout_pane_content_pbytes
} layout_pane_content_e;

/*
 * open console behaviour (win32 only)
 */
typedef enum {
    console_open_never,
    console_open_auto,
    console_open_always
} console_open_e;

/*
 * Places version information will show up
 */
typedef enum {
    version_welcome_only,
    version_title_only,
    version_both,
    version_neither
} version_info_e;

typedef enum {
    pref_default,
    pref_stashed,
    pref_current
} pref_source_t;

typedef enum {
    ELIDE_LEFT,
    ELIDE_RIGHT,
    ELIDE_MIDDLE,
    ELIDE_NONE
} elide_mode_e;


/*
 * Update channel.
 */
typedef enum {
    UPDATE_CHANNEL_DEVELOPMENT,
    UPDATE_CHANNEL_STABLE
} software_update_channel_e;

typedef struct _e_prefs {
  gint         pr_format;
  gint         pr_dest;
  gchar       *pr_file;
  gchar       *pr_cmd;
  GList       *col_list;
  gint         num_cols;
  color_t      st_client_fg, st_client_bg, st_server_fg, st_server_bg;
  color_t      gui_text_valid, gui_text_invalid, gui_text_deprecated;
  gboolean     gui_altern_colors; /* GTK only */
  gboolean     gui_expert_composite_eyecandy;
  gboolean     filter_toolbar_show_in_statusbar;
  gboolean     restore_filter_after_following_stream;
  gint         gui_ptree_line_style;
  gint         gui_ptree_expander_style;
  gboolean     gui_hex_dump_highlight_style;
  gint         gui_toolbar_main_style;
  gint         gui_toolbar_filter_style; /* GTK only? */
  gchar       *gui_gtk2_font_name;
  gchar       *gui_qt_font_name;
  color_t      gui_marked_fg;
  color_t      gui_marked_bg;
  color_t      gui_ignored_fg;
  color_t      gui_ignored_bg;
  gchar       *gui_colorized_fg;
  gchar       *gui_colorized_bg;
  gboolean     gui_geometry_save_position;
  gboolean     gui_geometry_save_size;
  gboolean     gui_geometry_save_maximized;
  gboolean     gui_macosx_style; /* GTK only */
  console_open_e gui_console_open;
  guint        gui_recent_df_entries_max;
  guint        gui_recent_files_count_max;
  guint        gui_fileopen_style;
  gchar       *gui_fileopen_dir;
  guint        gui_fileopen_preview;
  gboolean     gui_ask_unsaved;
  gboolean     gui_find_wrap;
  gboolean     gui_use_pref_save;
  gchar       *gui_webbrowser;
  gchar       *gui_window_title;
  gchar       *gui_prepend_window_title;
  gchar       *gui_start_title;
  version_info_e gui_version_placement;
  gboolean     gui_auto_scroll_on_expand; /* GTK+ only */
  guint        gui_auto_scroll_percentage; /* GTK+ only */
  layout_type_e gui_layout_type;
  layout_pane_content_e gui_layout_content_1;
  layout_pane_content_e gui_layout_content_2;
  layout_pane_content_e gui_layout_content_3;
  gchar       *gui_interfaces_hide_types;
  gboolean     gui_interfaces_show_hidden;
  gboolean     gui_interfaces_remote_display;
  gint         console_log_level;
  gchar       *capture_device;
  gchar       *capture_devices_linktypes;
  gchar       *capture_devices_descr;
  gchar       *capture_devices_hide;
  gchar       *capture_devices_monitor_mode;
  gchar       *capture_devices_buffersize;
  gchar       *capture_devices_snaplen;
  gchar       *capture_devices_pmode;
  gchar       *capture_devices_filter; /* XXX - Mostly unused. Deprecate? */
  gboolean     capture_prom_mode;
  gboolean     capture_pcap_ng;
  gboolean     capture_real_time;
  gboolean     capture_auto_scroll;
  gboolean     capture_no_extcap;
  gboolean     capture_show_info;
  GList       *capture_columns;
  guint        rtp_player_max_visible;
  guint        tap_update_interval;
  gboolean     display_hidden_proto_items;
  gboolean     display_byte_fields_with_spaces;
  gboolean     enable_incomplete_dissectors_check;
  gboolean     incomplete_dissectors_check_debug;
  gboolean     strict_conversation_tracking_heuristics;
  gboolean     filter_expressions_old;  /* TRUE if old filter expressions preferences were loaded. */
  gboolean     gui_update_enabled;
  software_update_channel_e gui_update_channel;
  gint         gui_update_interval;
  gchar       *saved_at_version;
  gboolean     unknown_prefs; /* unknown or obsolete pref(s) */
  gboolean     unknown_colorfilters; /* Warn when saving unknown or obsolete color filters. */
  gboolean     gui_qt_packet_list_separator;
  gboolean     gui_qt_show_selected_packet;
  gboolean     gui_qt_show_file_load_time;
  gboolean     gui_packet_editor; /* Enable Packet Editor */
  elide_mode_e gui_packet_list_elide_mode;
  gboolean     gui_packet_list_show_related;
  gboolean     gui_packet_list_show_minimap;
  gboolean     st_enable_burstinfo;
  gboolean     st_burst_showcount;
  gint         st_burst_resolution;
  gint         st_burst_windowlen;
  gboolean     st_sort_casesensitve;
  gboolean     st_sort_rng_fixorder;
  gboolean     st_sort_rng_nameonly;
  gint         st_sort_defcolflag;
  gboolean     st_sort_defdescending;
  gboolean     st_sort_showfullname;
  gboolean     extcap_save_on_start;
} e_prefs;

extern e_prefs prefs;

/*
 * Routines to let modules that have preference settings register
 * themselves by name, and to let them register preference settings
 * by name.
 */
struct pref_module;

struct pref_custom_cbs;

typedef struct pref_module module_t;

/** Sets up memory used by proto routines. Called at program startup */
/* FIXME: Not required void prefs_init(void);
*/

/** Reset preferences to default values.  Called at profile change */
extern void prefs_reset(void);

/** Frees memory used by proto routines. Called at program shutdown */
// Not required void prefs_cleanup(void);

/** Provide a hint about the darkness of the current UI theme so that we can adjust colors when needed */
extern void prefs_set_gui_theme_is_dark(gboolean is_dark);

/*
 * Register that a protocol has preferences.
 */
extern module_t *prefs_register_protocol(int id, void (*apply_cb)(void));

/**
 * Deregister preferences from a protocol.
 */
/* FIXME: Not exported
void prefs_deregister_protocol(int id);
*/

/*
 * Register that a statistical tap has preferences.
 *
 * "name" is a name for the tap to use on the command line with "-o"
 * and in preference files.
 *
 * "title" is a short human-readable name for the tap.
 *
 * "description" is a longer human-readable description of the tap.
 */
extern module_t *prefs_register_stat(const char *name, const char *title,
    const char *description, void (*apply_cb)(void));

/*
 * Register that a codec has preferences.
 *
 * "name" is a name for the codec to use on the command line with "-o"
 * and in preference files.
 *
 * "title" is a short human-readable name for the codec.
 *
 * "description" is a longer human-readable description of the codec.
 */
extern module_t *prefs_register_codec(const char *name, const char *title,
    const char *description, void (*apply_cb)(void));

/*
 * Register that a protocol has preferences and group it under a single
 * subtree
 */
// FIXME: Later -- #define PREFERENCE_GROUPING
extern module_t *prefs_register_protocol_subtree(const char *subtree, int id,
    void (*apply_cb)(void));

/*
 * Register that a protocol used to have preferences but no longer does,
 * by creating an "obsolete" module for it.
 */
/* FIXME: Not required
module_t *prefs_register_protocol_obsolete(int id);
*/

/*
 * Callback function for module list scanners.
 */
typedef guint (*module_cb)(module_t *module, gpointer user_data);

/*
 * Returns TRUE if module has any submodules
 */
extern gboolean prefs_module_has_submodules(module_t *module);

/*
 * Call a callback function, with a specified argument, for each module
 * in the list of all modules.  (This list does not include subtrees.)
 *
 * Ignores "obsolete" modules; their sole purpose is to allow old
 * preferences for dissectors that no longer have preferences to be
 * silently ignored in preference files.
 */
extern guint prefs_modules_foreach(module_cb callback, gpointer user_data);

/*
 * Call a callback function, with a specified argument, for each submodule
 * of specified modules.  If the module is NULL, goes through the top-level
 * list in the display tree of modules.
 *
 * Ignores "obsolete" modules; their sole purpose is to allow old
 * preferences for dissectors that no longer have preferences to be
 * silently ignored in preference files.  Does not ignore subtrees,
 * as this can be used when walking the display tree of modules.
 */
extern guint prefs_modules_foreach_submodules(module_t *module, module_cb callback, gpointer user_data);

/*
 * Call the "apply" callback function for each module if any of its
 * preferences have changed, and then clear the flag saying its
 * preferences have changed, as the module has been notified of that
 * fact.
 */
extern void prefs_apply_all(void);

/*
 * Call the "apply" callback function for a specific module if any of
 * its preferences have changed, and then clear the flag saying its
 * preferences have changed, as the module has been notified of that
 * fact.
 */
extern void prefs_apply(module_t *module);


struct preference;

typedef struct preference pref_t;

/*
 * Returns TRUE if the given protocol has registered preferences.
 */
extern gboolean prefs_is_registered_protocol(const char *name);

/*
 * Returns the module title of a registered protocol (or NULL if unknown).
 */
extern const char *prefs_get_title_by_name(const char *name);

/** Given a module name, return a pointer to its pref_module struct,
 * or NULL if it's not found.
 *
 * @param name The preference module name.  Usually the same as the protocol
 * name, e.g. "tcp".
 * @return A pointer to the corresponding preference module, or NULL if it
 * wasn't found.
 */
extern module_t *prefs_find_module(const char *name);

/** Given a module name, and a preference name return a pointer to the given
 * module's given preference or NULL if it's not found.
 *
 * @param module The preference module name.  Usually the same as the protocol
 * name, e.g. "tcp".
 * @param pref The preference name, e.g. "desegment".
 * @return A pointer to the corresponding preference, or NULL if it
 * wasn't found.
 */
extern pref_t *prefs_find_preference(module_t * module, const char *pref);

/*
 * Register a preference with an unsigned integral value.
 */
extern void prefs_register_uint_preference(module_t *module, const char *name,
    const char *title, const char *description, guint base, guint *var);

/*
 * prefs_register_ callers must conform to the following:
 *
 * Names must be in lowercase letters only (underscore allowed).
 * Titles and descriptions must be valid UTF-8 or NULL.
 * Titles must be short (less than 80 characters)
 * Titles must not contain newlines.
 */

/*
 * Register a preference with an Boolean value.
 */
extern void prefs_register_bool_preference(module_t *module, const char *name,
    const char *title, const char *description, gboolean *var);

/*
 * Register a preference with an enumerated value.
 */
extern void prefs_register_enum_preference(module_t *module, const char *name,
    const char *title, const char *description, gint *var,
    const enum_val_t *enumvals, gboolean radio_buttons);

/*
 * Register a preference with a character-string value.
 */
extern void prefs_register_string_preference(module_t *module, const char *name,
    const char *title, const char *description, const char **var);

/*
 * Register a preference with a file name (string) value.
 * File name preferences are basically like string preferences
 * except that the GUI gives the user the ability to browse for the
 * file.
 */
extern void prefs_register_filename_preference(module_t *module, const char *name,
    const char *title, const char *description, const char **var, gboolean for_writing);

/*
 * Register a preference with a directory name (string) value.
 * Directory name preferences are basically like string preferences
 * except that the GUI gives the user the ability to browse for a
 * directory.
 */
extern void prefs_register_directory_preference(module_t *module, const char *name,
    const char *title, const char *description, const char **var);

/*
 * Register a preference with a ranged value.
 */
extern void prefs_register_range_preference(module_t *module, const char *name,
    const char *title, const char *description, range_t **var,
    guint32 max_value);

/*
 * Register a static text 'preference'. It can be used to add some info/explanation.
 */
extern void prefs_register_static_text_preference(module_t *module, const char *name,
    const char *title, const char *description);

/*
 * Register a uat 'preference'. It adds a button that opens the uat's window in the
 * preferences tab of the module.
 */
extern void prefs_register_uat_preference(module_t *module,
    const char *name, const char* title, const char *description,  struct epan_uat* uat);

/*
 * Register a uat 'preference' for QT only. It adds a button that opens the uat's window in the
 * preferences tab of the module.
 */
extern void prefs_register_uat_preference_qt(module_t *module,
    const char *name, const char* title, const char *description,  struct epan_uat* uat);


/*
 * Register a color preference.  Currently does not have any "GUI Dialog" support
 * so the color data needs to be managed independently.  Currently used by the
 * "GUI preferences" to aid in reading/writing the preferences file, but the
 * "data" is still managed by the specific "GUI preferences" dialog.
 */
/* FIXME:
void prefs_register_color_preference(module_t *module, const char *name,
    const char *title, const char *description, color_t *color);
*/

/*
 * Register a custom preference.  Currently does not have any "GUI Dialog" support
 * so data needs to be managed independently.  Currently used by the
 * "GUI preferences" to aid in reading/writing the preferences file, but the
 * "data" is still managed by the specific "GUI preferences" dialog.
 */
/* FIXME: Not required
void prefs_register_custom_preference(module_t *module, const char *name,
    const char *title, const char *description, struct pref_custom_cbs* custom_cbs,
    void** custom_data);
*/

/*
 * Register a (internal) "Decode As" preference with a ranged value.
 */
/* FIXME: Not exported
void prefs_register_decode_as_range_preference(module_t *module, const char *name,
    const char *title, const char *description, range_t **var,
    guint32 max_value);
*/
/*
 * Register a (internal) "Decode As" preference with an unsigned integral value
 * for a dissector table.
 */
/* FIXME : Not Exported
void prefs_register_decode_as_preference(module_t *module, const char *name,
    const char *title, const char *description, guint *var);
*/

/*
 * Register a preference that used to be supported but no longer is.
 */
extern void prefs_register_obsolete_preference(module_t *module,
    const char *name);


typedef guint (*pref_cb)(pref_t *pref, gpointer user_data);

/*
 * Call a callback function, with a specified argument, for each preference
 * in a given module.
 *
 * If any of the callbacks return a non-zero value, stop and return that
 * value, otherwise return 0.
 */
extern guint prefs_pref_foreach(module_t *module, pref_cb callback,
    gpointer user_data);

/* Parse through a list of comma-separated, possibly quoted strings.
 *  Return a list of the string data.
 */
extern GList *prefs_get_string_list(const gchar *str);

/* Clear the given list of string data. */
extern void prefs_clear_string_list(GList *sl);

/** Fetch a short preference type name, e.g. "Integer".
 *
 * @param pref A preference.
 *
 * @return The preference type name. May be NULL.
 */
extern
const char *prefs_pref_type_name(pref_t *pref);

/** Fetch a long description of the preference type
 *
 * @param pref A preference.
 *
 * @return A description of the preference type including allowed
 * values for enums. The description may include newlines. Must be
 * g_free()d.
 */
extern
char *prefs_pref_type_description(pref_t *pref);

/** Fetch a string representation of the preference.
 *
 * @param pref A preference.
 * @param source Which value of the preference to return, see pref_source_t.
 *
 * @return A string representation of the preference. Must be g_free()d.
 */
extern
char *prefs_pref_to_str(pref_t *pref, pref_source_t source);

/* Read the preferences file, fill in "prefs", and return a pointer to it.

   If we got an error (other than "it doesn't exist") we report it through
   the UI.

   This is called by epan_load_settings(); programs should call that
   rather than individually calling the routines it calls. */
/* FIXME: Not exported
extern e_prefs *read_prefs(void);
*/

/* Write out "prefs" to the user's preferences file, and return 0.

   If we got an error, stuff a pointer to the path of the preferences file
   into "*pf_path_return", and return the errno. */
extern int write_prefs(char **);

/**
 * Result of setting a preference.
 */
typedef enum {
    PREFS_SET_OK,               /* succeeded */
    PREFS_SET_SYNTAX_ERR,       /* syntax error in string */
    PREFS_SET_NO_SUCH_PREF,     /* no such preference */
    PREFS_SET_OBSOLETE          /* preference used to exist but no longer does */
} prefs_set_pref_e;

/*
 * Given a string of the form "<pref name>:<pref value>", as might appear
 * as an argument to a "-o" option, parse it and set the preference in
 * question.  Return an indication of whether it succeeded or failed
 * in some fashion.
 *
 * For syntax errors (return value PREFS_SET_SYNTAX_ERR), details (when
 * available) are written into "errmsg" which must be freed with g_free.
 */
extern prefs_set_pref_e prefs_set_pref(char *prefarg, char **errmsg);

/*
 * Get or set a preference's obsolete status. These can be used to make a
 * preference obsolete after startup so that we can fetch its value but
 * keep it from showing up in the prefrences dialog.
 */
/*  FIXME:  Not required
gboolean prefs_get_preference_obsolete(pref_t *pref);
prefs_set_pref_e prefs_set_preference_obsolete(pref_t *pref);
*/

/*
 * Get current  preference uint value. This allows the preference structure
 * to remain hidden from those that doesn't really need it
 */
extern guint prefs_get_uint_value(const char *module_name, const char* pref_name);

/*
 * Get the current range preference value (maintained by pref, so it doesn't need to be freed). This allows the
 * preference structure to remain hidden from those that doesn't really need it.
 */
extern range_t* prefs_get_range_value(const char *module_name, const char* pref_name);

/*
 * Returns TRUE if the given device is hidden
 */
extern gboolean prefs_is_capture_device_hidden(const char *name);

/*
 * Returns TRUE if the given device should capture in monitor mode by default
 */
extern gboolean prefs_capture_device_monitor_mode(const char *name);

extern gboolean prefs_capture_options_dialog_column_is_visible(const gchar *column);

/*
 * Returns TRUE if the layout pane content is enabled
 */
extern gboolean prefs_has_layout_pane_content (layout_pane_content_e layout_pane_content);

/*
 * Editor modelines  -  http://www.wireshark.org/tools/modelines.html
 *
 * Local variables:
 * c-basic-offset: 4
 * tab-width: 8
 * indent-tabs-mode: nil
 * End:
 *
 * vi: set shiftwidth=4 tabstop=8 expandtab:
 * :indentSize=4:tabSize=8:noTabs=true:
 */

"""



epan_params_h_cdef = """

/* params.h
 * Definitions for parameter handling routines
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/*
 * Definition of a value for an enumerated type.
 *
 * "name" is the the name one would use on the command line for the value.
 * "description" is the description of the value, used in combo boxes/
 * option menus.
 * "value" is the value.
 */
typedef struct {
	const char	*name;
	const char	*description;
	gint		value;
} enum_val_t;



"""





glib_h_cdef = """

// From </usr/lib/x86_64-gnu-linux/gconfig.h>

typedef signed char gint8;
typedef unsigned char guint8;
typedef signed short gint16;
typedef unsigned short guint16;

typedef signed int gint32;
typedef unsigned int guint32;
#define G_HAVE_GINT64 1          /* deprecated, always true */

typedef signed long gint64;
typedef unsigned long guint64;

typedef signed long gssize;
typedef unsigned long gsize;

// From <glib-2.0/glib/gtypes.h>
typedef char   gchar;
typedef short  gshort;
typedef long   glong;
typedef int    gint;
typedef gint   gboolean;

typedef unsigned char   guchar;
typedef unsigned short  gushort;
typedef unsigned long   gulong;
typedef unsigned int    guint;

typedef float   gfloat;
typedef double  gdouble;

/* Define min and max constants for the fixed size numerical types
// Following are not allowed - Let's see how to fix this later

#define G_MININT8	((gint8) -0x80)
#define G_MAXINT8	((gint8)  0x7f)
#define G_MAXUINT8	((guint8) 0xff)

#define G_MININT16	((gint16) -0x8000)
#define G_MAXINT16	((gint16)  0x7fff)
#define G_MAXUINT16	((guint16) 0xffff)

#define G_MININT32	((gint32) -0x80000000)
#define G_MAXINT32	((gint32)  0x7fffffff)
#define G_MAXUINT32	((guint32) 0xffffffff)

#define G_MININT64	((gint64) G_GINT64_CONSTANT(-0x8000000000000000))
#define G_MAXINT64	G_GINT64_CONSTANT(0x7fffffffffffffff)
#define G_MAXUINT64	G_GUINT64_CONSTANT(0xffffffffffffffff)
*/
typedef void* gpointer;
typedef const void *gconstpointer;

// from <glib-2.0/glib/gslist.h>

typedef struct _GSList GSList;

struct _GSList
{
    gpointer data;
    GSList *next;
};

// from <glib-2.0/glib/gregex.h>
typedef struct _GRegex		GRegex;

typedef gint            (*GCompareFunc)         (gconstpointer  a,
                                                 gconstpointer  b);
typedef gint            (*GCompareDataFunc)     (gconstpointer  a,
                                                 gconstpointer  b,
						 gpointer       user_data);
typedef gboolean        (*GEqualFunc)           (gconstpointer  a,
                                                 gconstpointer  b);
typedef void            (*GDestroyNotify)       (gpointer       data);
typedef void            (*GFunc)                (gpointer       data,
                                                 gpointer       user_data);
typedef guint           (*GHashFunc)            (gconstpointer  key);
typedef void            (*GHFunc)               (gpointer       key,
                                                 gpointer       value,
                                                 gpointer       user_data);

"""



wsutil_plugins_h_cdef = """

/* plugins.h
 * definitions for plugins structures
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

typedef void plugins_t;

typedef enum {
    WS_PLUGIN_EPAN,
    WS_PLUGIN_WIRETAP,
    WS_PLUGIN_CODEC
} plugin_type_e;

extern plugins_t *plugins_init(plugin_type_e type);

typedef void (*plugin_description_callback)(const char *name, const char *version,
                                            const char *types, const char *filename,
                                            void *user_data);

extern void plugins_get_descriptions(plugin_description_callback callback, void *user_data);

extern void plugins_dump_all(void);

extern int plugins_get_count(void);

extern void plugins_cleanup(plugins_t *plugins);

"""



import os
import sys
import time
import struct
import socket
from datetime import datetime as dt
import warnings


try:
    from wishpy.wireshark.lib.epan3_ext import lib as epan_lib
    from wishpy.wireshark.lib.epan3_ext import ffi as epan_ffi
except ImportError:
    warnings.warn("Bindings for supported wireshark Version (3.2.x) Not found.")
    sys.exit(-1)

_MAX_TO_PROCESS = 10000000

# FIXME: This should come from some common utils
# Make sure we are indeed wireshark 3.2
major = epan_ffi.new('int *')
minor = epan_ffi.new('int *')
micro = epan_ffi.new('int *')

epan_lib.epan_get_version_number(major, minor, micro)
if major[0] != 3 and minor[0] != 2:
    version_str = "{}.{}.{}".format(major[0], minor[0], micro[0])
    warn_str = "This is supported only with Wireshark 3.2.x, found version {}".\
            format(version_str)
    warnings.warn(warn_str)
    sys.exit(1)


nstime_empty = epan_ffi.new('nstime_t *');

@epan_ffi.callback('const nstime_t *(*)(struct packet_provider_data *prov, guint32 frame_num)')
def wishpy_get_ts(prov, frame_num):

    return nstime_empty

wishpy_provider_funcs = epan_ffi.new('struct packet_provider_funcs *',
        [wishpy_get_ts, epan_ffi.NULL, epan_ffi.NULL, epan_ffi.NULL])

# Below are some dict's required for printing few packet types

hfbases = {
        epan_lib.BASE_NONE : '{:d}', # Not sure how this is to be treated
        epan_lib.BASE_DEC : '{:d}',
        epan_lib.BASE_HEX : '0x{:x}',
        epan_lib.BASE_OCT : '{:o}',
        epan_lib.BASE_DEC_HEX: '{:d}(0x{:x})',
        epan_lib.BASE_HEX_DEC: '0x{:x}({:d})',
        epan_lib.BASE_PT_TCP: '{:d}',
        epan_lib.BASE_PT_UDP: '{:d}',
        epan_lib.BASE_PT_SCTP: '{:d}'

        }
def func_not_supported(*args):
    return "Not Supported"

def epan_ether_to_str(fvalue, ftype, display):

    eth_bytes = fvalue.value.bytes

    display_bytes = []
    for i in range(eth_bytes.len):
        display_byte = "{:02X}".format(eth_bytes.data[i])
        display_bytes.append(display_byte)

    return ":".join(display_bytes)

epan_bytes_to_str = epan_ether_to_str

def epan_str_to_str(fvalue, ftype, display):

    value = fvalue.value.string
    return epan_ffi.string(value).decode()

def epan_ipv4_to_str(fvalue, ftype, display):

    ipv4 = fvalue.value.ipv4

    return socket.inet_ntoa(struct.pack('!I', ipv4.addr))

def epan_bool_to_str(fvalue, ftype, display, on_off=False, json_compat=True):

    if on_off and json_compat:
        raise ValueError("Specify Either on_off or json_compat not both.")

    value = bool(fvalue.value.uinteger)

    if value:
        if json_compat:
            return "true"
        if on_off:
            return "ON"
    else:
        if json_compat:
            return "false"
        if on_off:
            return "OFF"

    return "{}".format(value)


def epan_int_to_str(fvalue, ftype, display):

    try:
        # We are not displaying 'Extended string for the value
        if display & epan_lib.BASE_EXT_STRING:
            display ^= epan_lib.BASE_EXT_STRING

        base_format = hfbases[display]
    except:
        return "type: {} display: {} Not Supported".format(ftype, display)

    return base_format.format(fvalue.value.uinteger,
            fvalue.value.uinteger)

def value_to_str(finfo):
    """
    Returns string representation of the `finfo.value`
    """

    fvalue = finfo.value
    ftype = finfo.hfinfo[0].type
    display = finfo.hfinfo[0].display

    epan_int_types = [
            epan_lib.FT_INT8,
            epan_lib.FT_INT16,
            epan_lib.FT_INT32,
            epan_lib.FT_INT40,
            epan_lib.FT_INT48,
            epan_lib.FT_INT56,
            epan_lib.FT_INT64]

    epan_uint32_types = [
            epan_lib.FT_CHAR,
            epan_lib.FT_UINT8,
            epan_lib.FT_UINT16,
            epan_lib.FT_UINT32,
            epan_lib.FT_FRAMENUM]

    epan_uint_types = epan_uint32_types + \
            [ epan_lib.FT_UINT40, epan_lib.FT_UINT48,
            epan_lib.FT_UINT56, epan_lib.FT_UINT64]

    epan_all_int_types = epan_int_types + epan_uint_types

    if ftype in epan_all_int_types:
        return epan_int_to_str(fvalue, ftype, display)

    if ftype == epan_lib.FT_ETHER:
        return epan_ether_to_str(fvalue, ftype, display)

    if ftype == epan_lib.FT_IPv4:
        return epan_ipv4_to_str(fvalue, ftype, display)

    if ftype == epan_lib.FT_BOOLEAN:
        return epan_bool_to_str(fvalue, ftype, display)

    if ftype == epan_lib.FT_STRING:
        return epan_str_to_str(fvalue, ftype, display)

    if ftype == epan_lib.FT_BYTES:
        return epan_bytes_to_str(fvalue, ftype, display)

    if ftype in [epan_lib.FT_NONE, epan_lib.FT_PROTOCOL]:
        return ""

    return "{} {}".format(ftype, display)

def print_dissected_tree(node_ptr, data_ptr):

    return_str = ""

    if data_ptr == epan_ffi.NULL:
        level = 1
    else:
        level = epan_ffi.cast('int *', data_ptr)[0]

    return_str = ""
    node = node_ptr[0]
    finfo = node.finfo
    if finfo != epan_ffi.NULL:

        hfinfo = finfo.hfinfo[0]
        abbrev = epan_ffi.string(hfinfo.abbrev).decode()
        abbrev_str = "\"{!s}\"".format(abbrev)
        return_str += abbrev_str + ": "
        finfo_display_str = value_to_str(finfo)
        if finfo_display_str:
            finfo_display_str = "\"{!s}\"".format(finfo_display_str)
    else:
        finfo_display_str = ""

    return_str += finfo_display_str

    data_ptr_new = epan_ffi.new('int *')
    data_ptr_new[0] = level + 1
    child = node.first_child
    if child != epan_ffi.NULL:

        if finfo_display_str:
            return_str += ",\n"

            abbrev_tree = abbrev + "_tree"
            abbrev_tree_str = "\"{!s}\"".format(abbrev_tree)
            return_str += "  " * (level -1)
            return_str += abbrev_tree_str + " : "

        return_str += "{ "
        return_str += "\n"
        while child != epan_ffi.NULL:
            return_str += "  " * level
            return_str += print_dissected_tree(child, data_ptr_new)
            child = child.next
        return_str += "  " * level

        return_str += "\n"
        return_str += "  " * (level-1)
        return_str += "}"
    if node.next != epan_ffi.NULL:
        return_str += ",\n"

    return return_str

def packet_to_json(handle_ptr):

    dissector = handle_ptr[0]
    print(print_dissected_tree(dissector.tree, epan_ffi.NULL))

def wtap_open_file_offline(filepath):
    """
    Opens a file given by filepath  using `wtap_open_offline` and returns a tuple containing
    Handle and file type.
    """

    if not os.path.exists(filepath):
        return None, None


    epan_lib.wtap_init(False)

    # Open a wtap file
    err = epan_ffi.new('int *')
    err_str = epan_ffi.new("gchar *[256]")
    filename = filepath
    open_type = epan_lib.WTAP_TYPE_AUTO
    wth = epan_lib.wtap_open_offline(filename.encode(), open_type, err, err_str, False)
    if wth is epan_ffi.NULL:
        return None, None

    wtap_file_type = epan_lib.wtap_file_type_subtype(wth)
    return wth, wtap_file_type

# Open `epan_lib`

def epan_lib_init():
    """
    Performs initialzation of the `epan` library. Calls
    1. `init_process_policies` (Required)
    2. `epan_init` (Required)
    3. `epan_load_settings` (Required)

    Returns True/False based upon result of operation.
    """

    epan_lib.init_process_policies()
    null_register_cb = epan_ffi.cast('register_cb', epan_ffi.NULL)
    result = epan_lib.epan_init(
            null_register_cb, epan_ffi.NULL, True)
    if result:
        epan_lib.epan_load_settings()

    return result

def epan_perform_dissection(wth, wth_file_type):
    """
    Performs dissection of the file bound to `wth`
    """
    #empty_packet_provider_funcs = epan_ffi.new('struct packet_provider_funcs *')
    epan_session = epan_lib.epan_new(epan_ffi.NULL, wishpy_provider_funcs)

    # TODO: make proper `epan_session` for us
    epan_dissect_obj = epan_lib.epan_dissect_new(epan_session, True, True)

    offset = epan_ffi.new('gint64 *')
    frame_data_ref = epan_ffi.new('frame_data **')
    elapsed_time_ptr = epan_ffi.new('nstime_t *')

    buf = epan_ffi.new('Buffer *')
    epan_lib.ws_buffer_init(buf, 1514) # FIXME : Should do with proper length
    rec = epan_ffi.new('wtap_rec *')
    epan_lib.wtap_rec_init(rec)
    err = epan_ffi.new('int *')
    err_str = epan_ffi.new("gchar **")

    frame_data_ref[0] = epan_ffi.NULL
    frame_data_ptr = epan_ffi.new('frame_data *')
    cum_bytes = epan_ffi.new('guint32 *')
    then = dt.now()
    processed = 0
    total_bytes = 0
    while True:
        result = epan_lib.wtap_read(wth, rec, buf, err, err_str, offset)

        if result == True:
            processed += 1

            pkt_len = rec.rec_header.packet_header.len
            pkt_reported_len = rec.rec_header.packet_header.caplen


            epan_lib.frame_data_init(frame_data_ptr, processed, rec,
                    offset[0], cum_bytes[0])

            total_bytes += pkt_reported_len
            cum_bytes[0] = total_bytes
            # FIXME: Look at properly using `frame_data_ref`
            epan_lib.frame_data_set_before_dissect(frame_data_ptr,
                    elapsed_time_ptr, frame_data_ref, epan_ffi.NULL)


            # Not sure what this is - Just copied from `tshark` sources
            epan_lib.prime_epan_dissect_with_postdissector_wanted_hfids(
                    epan_dissect_obj)

            ## Do actual dissection here.
            # Get buffer and tvbuff first and then run dissector
            tvb_ptr = epan_lib.tvb_new_real_data(buf[0].data, pkt_len,
                    pkt_reported_len)

            epan_lib.epan_dissect_run(epan_dissect_obj, wth_file_type,
                    rec, tvb_ptr, frame_data_ptr, epan_ffi.NULL)

            packet_to_json(frame_data_ptr, epan_dissect_obj)

            # Reset the frame data and dissector object
            epan_lib.frame_data_set_after_dissect(frame_data_ptr, cum_bytes)
            epan_lib.epan_dissect_reset(epan_dissect_obj)

            processed += 1

            if processed == _MAX_TO_PROCESS:
                break

        else:
            break

    now = dt.now()

    print("Processded {} bytes from {} packets in {}".format(
            total_bytes, processed, (now - then)))


if __name__ == '__main__':

    if not len(sys.argv) >= 2:
        print("Usage: tshark.py <filepath>")
        sys.exit(1)

    input_filepath = sys.argv[1]

    wth, wth_ftype = wtap_open_file_offline(input_filepath)
    if wth is None:
        print("Unable to get 'wiretap' File Handle.")
        sys.exit(1)

    epan_lib_result = epan_lib_init()
    if not epan_lib_result:
        print("Unable to open 'epan' library.")
        sys.exit(1)

    epan_perform_dissection(wth, wth_ftype)



wsutil_nstime_h_types_cdef = """

/* nstime.h
 * Definition of data structure to hold time values with nanosecond resolution
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

// stuff from C standard library without having to do lot of including
typedef long int time_t; // Taken from /usr/include/x86_64

// From <wsutil/nstime.h>

typedef struct {
	time_t	secs;
	int	nsecs;
} nstime_t;



"""

wsutil_nstime_h_funcs_cdef = """


/* functions */

/** set the given nstime_t to zero */
extern void nstime_set_zero(nstime_t *nstime);

/** is the given nstime_t currently zero? */
extern gboolean nstime_is_zero(nstime_t *nstime);

/** set the given nstime_t to (0,maxint) to mark it as "unset"
 * That way we can find the first frame even when a timestamp
 * is zero (fix for bug 1056)
 */
extern void nstime_set_unset(nstime_t *nstime);

/* is the given nstime_t currently (0,maxint)? */
extern gboolean nstime_is_unset(const nstime_t *nstime);

/** duplicate the current time
 *
 * a = b
 */
extern void nstime_copy(nstime_t *a, const nstime_t *b);

/** calculate the delta between two times (can be negative!)
 *
 * delta = b-a
 *
 * Note that it is acceptable for two or more of the arguments to point at the
 * same structure.
 */
extern void nstime_delta(nstime_t *delta, const nstime_t *b, const nstime_t *a );

/** calculate the sum of two times
 *
 * sum = a+b
 *
 * Note that it is acceptable for two or more of the arguments to point at the
 * same structure.
 */
extern void nstime_sum(nstime_t *sum, const nstime_t *b, const nstime_t *a );

/** sum += a */
#define nstime_add(sum, a) nstime_sum(sum, sum, a)

/** sum -= a */
#define nstime_subtract(sum, a) nstime_delta(sum, sum, a)

/** compare two times are return a value similar to memcmp() or strcmp().
 *
 * a > b : > 0
 * a = b : 0
 * a < b : < 0
 */
extern int nstime_cmp (const nstime_t *a, const nstime_t *b );

/** converts nstime to double, time base is milli seconds */
extern double nstime_to_msec(const nstime_t *nstime);

/** converts nstime to double, time base is seconds */
extern double nstime_to_sec(const nstime_t *nstime);

/** converts Windows FILETIME to nstime, returns TRUE on success,
    FALSE on failure */
extern gboolean filetime_to_nstime(nstime_t *nstime, guint64 filetime);

/** converts time like Windows FILETIME, but expressed in nanoseconds
    rather than tenths of microseconds, to nstime, returns TRUE on success,
    FALSE on failure */
extern gboolean nsfiletime_to_nstime(nstime_t *nstime, guint64 nsfiletime);

"""



from wishpy.wireshark.src.wireshark3.epan.epan import (
        epan_ffi,
        _sources,
        _libraries,
        _extra_compile_args,
        _extra_link_args)

if __name__ == '__main__':
    epan_lib = epan_ffi.verify(_sources,
            libraries=_libraries,
            extra_compile_args=_extra_compile_args,
            extra_link_args=_extra_link_args)

    print(epan_lib)



import os
import sys
import subprocess
import warnings
from setuptools import setup, find_packages

def find_libwireshark_version():

    # First try using pkg-config
    major, minor, patch = None, None, None
    try:
        output = subprocess.run(["pkg-config", "--modversion", "wireshark"],
                check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        major, minor, _ = output.stdout.decode().strip().split(".")
        major, minor = int(major), int(minor)
        return major, minor, True
    except subprocess.CalledProcessError:
        pass


    if os.path.exists("wireshark-version"):
        with open("wireshark-version", "r") as f:
            for line in f:
                if line.lower().startswith("version:"):
                    version_str = line.split(":")[1].strip()
                    major, minor, _ = version_str.split(".")
                    major, minor = int(major), int(minor)
                    return major, minor, False

    return None, None, None

major, minor, is_pkgconfig = find_libwireshark_version()

if major is None or minor is None:
    major = 2
    minor = 6

version_str = "{}.{}".format(major, minor)

if not is_pkgconfig:
    warning_msg = "Heuristically determined version of Wireshark is {}.{} .".\
            format(major, minor)
    warning_msg += " Trying to find libs/includes in the prefix '/usr/local/"
    warnings.warn(warning_msg)

if major == 2:
    if minor != 6:
        print("Supported version is 2.6. Determined version is ", version_str)
        sys.exit(-1)

    epan_ffi_module = 'wishpy/wireshark/src/wireshark2/epan/epan_builder.py:epan_ffi'

elif major == 3:
    if minor != 2:
        print("Supported version is 3.2. Determined version is ", version_str)
        sys.exit(-1)

    epan_ffi_module = 'wishpy/wireshark/src/wireshark3/epan/epan_builder.py:epan_ffi'

libpcap_ffi_module = 'wishpy/libpcap/src/pcap_builder.py:libpcap_ffi'

setup(name='wishpy',
        version='0.0.7',
        description='Python Bindings for Wireshark using cffi',
        author='hyphenOs Software Labs',
        author_email='gabhijit@hyphenos.io',
        license_files=['LICENSE', 'COPYING', 'COPYING-Wireshark', 'LICENSE-libpcap'],
        setup_requires=['cffi>=1.14.0'],
        install_requires=['cffi>=1.14.0'],
        cffi_modules=[
            epan_ffi_module,
            libpcap_ffi_module
            ],
        packages=find_packages(),
        scripts=['examples/tshark3.py',
            'examples/tshark2.py',
            'examples/tshark.py',
            'examples/tcpdump.py'],
        zip_safe=False)



from cffi import FFI
try:
    from cffi import PkgConfigError
except:
    pass

from ...glib.glib_h import glib_h_cdef
from ...glib.garray_h import garray_h_cdef
from ...glib.glist_h import glist_h_cdef
from ...glib.gstring_h import glib_gstring_h_types_cdef
from ...glib.ghash_h import glib_ghash_h_types_cdef

from ...wsutil.nstime_h import wsutil_nstime_h_types_cdef
from ...wsutil.buffer_h import wsutil_buffer_h_cdef
from ...wsutil.inet_ipv4_h import wsutil_inet_ipv4_h_cdef
from ...wsutil.inet_ipv6_h import wsutil_inet_ipv6_h_cdef
from ...wsutil.inet_addr_h import wsutil_inet_addr_h_cdef
from ...wsutil.plugins_h import wsutil_plugins_h_cdef
from ...wsutil.colors_h import wsutil_colors_h_cdef
#from ...wsutil.ws_mempbrk_h import wsutil_ws_mempbrk_h_types_cdef
from ...wsutil.privileges_h import wsutil_privileges_h_funcs_cdef


from ..wtap.wtap_h import (
        wtap_h_types_cdef,
        wtap_h_funcs_cdef)

from ..wtap.wtap_opttypes_h import (
        wtap_opttypes_h_types_cdef,
        wtap_opttypes_h_funcs_cdef)

from .epan_h import epan_h_cdef
from .register_h import epan_register_h_types_cdef
from .framedata_h import (
        framedata_h_types_cdef,
        framedata_h_funcs_cdef
        )
from .prefs_h import epan_prefs_h_cdef
from .params_h import epan_params_h_cdef
from .range_h import epan_range_h_types_cdef
from .tvbuff_h import epan_tvbuff_h_types_cdef
from .tvbuff_h import epan_tvbuff_h_funcs_cdef
from .guid_utils_h import epan_guid_utils_h_types_cdef
from .wmem_h import epan_wmem_h_types_cdef
from .packet_h import epan_packet_h_funcs_cdef, epan_packet_h_types_cdef
from .address_h import epan_address_h_types_cdef
from .packet_info_h import epan_packet_info_h_cdef
from .epan_dissect_h import epan_epan_dissect_h_types_cdef
from .proto_h import epan_proto_h_types_cdef
from .ftypes_h import epan_ftypes_h_types_cdef


epan_ffi = FFI()

epan_ffi.cdef(glib_h_cdef)
epan_ffi.cdef(garray_h_cdef)
epan_ffi.cdef(glist_h_cdef)
epan_ffi.cdef(glib_gstring_h_types_cdef)
epan_ffi.cdef(glib_ghash_h_types_cdef)

epan_ffi.cdef(wsutil_nstime_h_types_cdef)
epan_ffi.cdef(wsutil_buffer_h_cdef)
epan_ffi.cdef(wsutil_inet_ipv4_h_cdef)
epan_ffi.cdef(wsutil_inet_ipv6_h_cdef)
epan_ffi.cdef(wsutil_inet_addr_h_cdef)
epan_ffi.cdef(wsutil_plugins_h_cdef)
epan_ffi.cdef(wsutil_colors_h_cdef)
epan_ffi.cdef(wsutil_privileges_h_funcs_cdef)

epan_ffi.cdef(wtap_opttypes_h_types_cdef)
epan_ffi.cdef(wtap_opttypes_h_funcs_cdef)
epan_ffi.cdef(wtap_h_types_cdef)
epan_ffi.cdef(wtap_h_funcs_cdef)

epan_ffi.cdef(framedata_h_types_cdef)
epan_ffi.cdef(framedata_h_funcs_cdef)
epan_ffi.cdef(epan_register_h_types_cdef)
epan_ffi.cdef(epan_params_h_cdef)
epan_ffi.cdef(epan_range_h_types_cdef)
epan_ffi.cdef(epan_prefs_h_cdef)

#epan_ffi.cdef(wsutil_ws_mempbrk_h_types_cdef)
epan_ffi.cdef(epan_guid_utils_h_types_cdef)
epan_ffi.cdef(epan_wmem_h_types_cdef)
epan_ffi.cdef(epan_tvbuff_h_types_cdef)
epan_ffi.cdef(epan_tvbuff_h_funcs_cdef)
epan_ffi.cdef(epan_address_h_types_cdef)
epan_ffi.cdef(epan_packet_info_h_cdef)
epan_ffi.cdef(epan_ftypes_h_types_cdef)
epan_ffi.cdef(epan_proto_h_types_cdef)
epan_ffi.cdef(epan_epan_dissect_h_types_cdef)
epan_ffi.cdef(epan_h_cdef)
epan_ffi.cdef(epan_packet_h_types_cdef)
epan_ffi.cdef(epan_packet_h_funcs_cdef)

# Go ahead and get the Library handle


_sources = '''
            #define HAVE_PLUGINS 1
            #include <epan/address.h>
            #include <epan/proto.h>
            #include <epan/epan_dissect.h>
            #include <epan/epan.h>
            #include <epan/packet.h>
            #include <epan/prefs.h>
            #include <wsutil/privileges.h>
'''

_pkg_name = 'wishpy.wireshark.lib.epan3_ext'

_pkgconfig_libs = ['wireshark']

_libraries = ['glib-2.0', 'wireshark', 'wsutil']

_extra_compile_args = [
        '-I/usr/local/include/wireshark',
        '-I/usr/include/wireshark',
        '-I/usr/include/glib-2.0',
        '-I/usr/lib/x86_64-linux-gnu/glib-2.0/include',
        ]

_extra_link_args = [
        '-L/usr/local/lib',
        ]

try:
    epan_ffi.set_source_pkgconfig(_pkg_name, _pkgconfig_libs, _sources,
            extra_link_args=_extra_link_args)
except PkgConfigError:
    epan_ffi._set_source(_pkg_name, _sources,
            libraries=_libraries,
            extra_compile_args=_extra_compile_args,
            extra_link_args=_extra_link_args)




epan_h_cdef = """


typedef struct epan_dissect epan_dissect_t;

struct epan_dfilter;
struct epan_column_info;

/*
 * Opaque structure provided when an epan_t is created; it contains
 * information needed to allow the user of libwireshark to provide
 * time stamps, comments, and other information outside the packet
 * data itself.
 */
struct packet_provider_data;

/*
 * Structure containing pointers to functions supplied by the user
 * of libwireshark.
 */
struct packet_provider_funcs {
	const nstime_t *(*get_frame_ts)(struct packet_provider_data *prov, guint32 frame_num);
	const char *(*get_interface_name)(struct packet_provider_data *prov, guint32 interface_id);
	const char *(*get_interface_description)(struct packet_provider_data *prov, guint32 interface_id);
	const char *(*get_user_comment)(struct packet_provider_data *prov, const frame_data *fd);
};

// FIXME : perhaps not required extern plugins_t *libwireshark_plugins;

/**
	@mainpage Wireshark EPAN the packet analyzing engine. Source code can be found in the epan directory

	@section Introduction

	XXX

	@b Sections:
*/
/*
Ref 1
Epan
Ethereal Packet ANalyzer (XXX - is this correct?) the packet analyzing engine. Source code can be found in the epan directory.

Protocol-Tree - Keep data of the capture file protocol information.

Dissectors - The various protocol dissectors in epan/dissectors.

Plugins - Some of the protocol dissectors are implemented as plugins. Source code can be found at plugins.

Display-Filters - the display filter engine at epan/dfilter



Ref2 for further edits - delete when done
	\section Introduction

	This document describes the data structures and the functions exported by the CACE Technologies AirPcap library.
	The AirPcap library provides low-level access to the AirPcap driver including advanced capabilities such as channel setting,
	link type control and WEP configuration.<br>
	This manual includes the following sections:

	\note throughout this documentation, \e device refers to a physical USB AirPcap device, while \e adapter is an open API
	instance. Most of the AirPcap API operations are adapter-specific but some of them, like setting the channel, are
	per-device and will be reflected on all the open adapters. These functions will have "Device" in their name, e.g.
	AirpcapSetDeviceChannel().

	\b Sections:

	- \ref airpcapfuncs
	- \ref airpcapdefs
	- \ref radiotap
*/

/**
 * Init the whole epan module.
 *
 * Must be called only once in a program.
 *
 * Returns TRUE on success, FALSE on failure.
 */
extern
gboolean epan_init(register_cb cb, void *client_data, gboolean epan_load_plugins);

/**
 * Load all settings, from the current profile, that affect epan.
 */
extern
e_prefs *epan_load_settings(void);

/** cleanup the whole epan module, this is used to be called only once in a program */
extern
void epan_cleanup(void);

typedef struct {
	void (*init)(void);
	void (*dissect_init)(epan_dissect_t *);
	void (*dissect_cleanup)(epan_dissect_t *);
	void (*cleanup)(void);
	void (*register_all_protocols)(register_cb, gpointer);
	void (*register_all_handoffs)(register_cb, gpointer);
        void (*register_all_tap_listeners)(void);
} epan_plugin;

extern void epan_register_plugin(const epan_plugin *plugin);
/**
 * Initialize the table of conversations.  Conversations are identified by
 * their endpoints; they are used for protocols such as IP, TCP, and UDP,
 * where packets contain endpoint information but don't contain a single
 * value indicating to which flow the packet belongs.
 */
/* FIXME: Not exported
void epan_conversation_init(void);
*/

/** A client will create one epan_t for an entire dissection session.
 * A single epan_t will be used to analyze the entire sequence of packets,
 * sequentially, in a single session. A session corresponds to a single
 * packet trace file. The reasons epan_t exists is that some packets in
 * some protocols cannot be decoded without knowledge of previous packets.
 * This inter-packet "state" is stored in the epan_t.
 */
typedef struct epan_session epan_t;

extern epan_t *epan_new(struct packet_provider_data *prov,
    const struct packet_provider_funcs *funcs);

extern const char *epan_get_user_comment(const epan_t *session, const frame_data *fd);

extern const char *epan_get_interface_name(const epan_t *session, guint32 interface_id);

extern const char *epan_get_interface_description(const epan_t *session, guint32 interface_id);

// FIXME : Not exported const nstime_t *epan_get_frame_ts(const epan_t *session, guint32 frame_num);

extern void epan_free(epan_t *session);

extern const gchar*
epan_get_version(void);

extern void epan_get_version_number(int *major, int *minor, int *micro);

/**
 * Set/unset the tree to always be visible when epan_dissect_init() is called.
 * This state change sticks until cleared, rather than being done per function call.
 * This is currently used when Lua scripts request all fields be generated.
 * By default it only becomes visible if epan_dissect_init() makes it so, usually
 * only when a packet is selected.
 * Setting this overrides that so it's always visible, although it will still not be
 * created if create_proto_tree is false in the call to epan_dissect_init().
 * Clearing this reverts the decision to epan_dissect_init() and proto_tree_visible.
 */
/* FIXME : Not required
void epan_set_always_visible(gboolean force);
*/

/** initialize an existing single packet dissection */
extern
void
epan_dissect_init(epan_dissect_t *edt, epan_t *session, const gboolean create_proto_tree, const gboolean proto_tree_visible);

/** get a new single packet dissection
 * should be freed using epan_dissect_free() after packet dissection completed
 */
extern
epan_dissect_t*
epan_dissect_new(epan_t *session, const gboolean create_proto_tree, const gboolean proto_tree_visible);

extern
void
epan_dissect_reset(epan_dissect_t *edt);

/** Indicate whether we should fake protocols or not */
extern
void
epan_dissect_fake_protocols(epan_dissect_t *edt, const gboolean fake_protocols);

/** run a single packet dissection */
extern
void
epan_dissect_run(epan_dissect_t *edt, int file_type_subtype,
        wtap_rec *rec, tvbuff_t *tvb, frame_data *fd,
        struct epan_column_info *cinfo);

extern
void
epan_dissect_run_with_taps(epan_dissect_t *edt, int file_type_subtype,
        wtap_rec *rec, tvbuff_t *tvb, frame_data *fd,
        struct epan_column_info *cinfo);

/** run a single file packet dissection */
extern
void
epan_dissect_file_run(epan_dissect_t *edt, wtap_rec *rec,
        tvbuff_t *tvb, frame_data *fd, struct epan_column_info *cinfo);

extern
void
epan_dissect_file_run_with_taps(epan_dissect_t *edt, wtap_rec *rec,
        tvbuff_t *tvb, frame_data *fd, struct epan_column_info *cinfo);

/** Prime an epan_dissect_t's proto_tree using the fields/protocols used in a dfilter. */
extern
void
epan_dissect_prime_with_dfilter(epan_dissect_t *edt, const struct epan_dfilter *dfcode);

/** Prime an epan_dissect_t's proto_tree with a field/protocol specified by its hfid */
extern
void
epan_dissect_prime_with_hfid(epan_dissect_t *edt, int hfid);

/** Prime an epan_dissect_t's proto_tree with a set of fields/protocols specified by their hfids in a GArray */
extern
void
epan_dissect_prime_with_hfid_array(epan_dissect_t *edt, GArray *hfids);

/** fill the dissect run output into the packet list columns */
extern
void
epan_dissect_fill_in_columns(epan_dissect_t *edt, const gboolean fill_col_exprs, const gboolean fill_fd_colums);

/** Check whether a dissected packet contains a given named field */
extern
gboolean
epan_dissect_packet_contains_field(epan_dissect_t* edt,
                                   const char *field_name);

/** releases resources attached to the packet dissection. DOES NOT free the actual pointer */
extern
void
epan_dissect_cleanup(epan_dissect_t* edt);

/** free a single packet dissection */
extern
void
epan_dissect_free(epan_dissect_t* edt);

/** Sets custom column */
/* FIXME : Not required
const gchar *
epan_custom_set(epan_dissect_t *edt, GSList *ids, gint occurrence,
				gchar *result, gchar *expr, const int size);
*/

/**
 * Get compile-time information for libraries used by libwireshark.
 */
extern
void
epan_get_compiled_version_info(GString *str);

/**
 * Get runtime information for libraries used by libwireshark.
 */
extern
void
epan_get_runtime_version_info(GString *str);

"""



# Types from wireshark/wtap/wtap_opttypes.h

wtap_opttypes_h_types_cdef = """

/* wtap_opttypes.h
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 2001 Gerald Combs
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/*
 * We use the pcapng option codes for option type values.
 */

/* Options for all blocks */
#define OPT_EOFOPT           0     /**< Appears in pcapng files, but not in blocks. */
#define OPT_COMMENT          1     /**< A UTF-8 string containing a human-readable comment. */

/* Section Header block (SHB) */
#define OPT_SHB_HARDWARE     2     /**< A UTF-8 string containing the description of the
                                     *     hardware used to create this section.
                                     */
#define OPT_SHB_OS           3     /**< A UTF-8 string containing the
                                     *     name of the operating system used to create this section.
                                     */
#define OPT_SHB_USERAPPL     4     /**< A UTF-8 string containing the
                                     *     name of the application used to create this section.
                                     */

/* Interface Description block (IDB) */
#define OPT_IDB_NAME         2     /**< A UTF-8 string containing the name
                                     *     of the device used to capture data.
                                     *     "eth0" / "\Device\PF_{AD1CE675-96D0-47C5-ADD0-2504B9126B68}"
                                     */
#define OPT_IDB_DESCR        3     /**< A UTF-8 string containing the description
                                     *     of the device used to capture data.
                                     *     "Wi-Fi" / "Local Area Connection" /
                                     *     "Wireless Network Connection" /
                                     *     "First Ethernet Interface"
                                     */
#define OPT_IDB_IP4ADDR      4     /**< XXX: if_IPv4addr Interface network address and netmask.
                                     *     This option can be repeated multiple times within the same Interface Description Block
                                     *     when multiple IPv4 addresses are assigned to the interface.
                                     *     192 168 1 1 255 255 255 0
                                     */
#define OPT_IDB_IP6ADDR      5     /* XXX: if_IPv6addr Interface network address and prefix length (stored in the last byte).
                                     *     This option can be repeated multiple times within the same Interface
                                     *     Description Block when multiple IPv6 addresses are assigned to the interface.
                                     *     2001:0db8:85a3:08d3:1319:8a2e:0370:7344/64 is written (in hex) as
                                     *     "20 01 0d b8 85 a3 08 d3 13 19 8a 2e 03 70 73 44 40"*/
#define OPT_IDB_MACADDR      6     /* XXX: if_MACaddr  Interface Hardware MAC address (48 bits).                             */
#define OPT_IDB_EUIADDR      7     /* XXX: if_EUIaddr  Interface Hardware EUI address (64 bits)                              */
#define OPT_IDB_SPEED        8     /**< Interface speed (in bps). 100000000 for 100Mbps
                                     */
#define OPT_IDB_TSRESOL      9     /**< Resolution of timestamps. If the Most Significant Bit is equal to zero,
                                     *     the remaining bits indicates the resolution of the timestamp as as a
                                     *     negative power of 10 (e.g. 6 means microsecond resolution, timestamps
                                     *     are the number of microseconds since 1/1/1970). If the Most Significant Bit
                                     *     is equal to one, the remaining bits indicates the resolution has a
                                     *     negative power of 2 (e.g. 10 means 1/1024 of second).
                                     *     If this option is not present, a resolution of 10^-6 is assumed
                                     *     (i.e. timestamps have the same resolution of the standard 'libpcap' timestamps).
                                     */
#define OPT_IDB_TZONE        10    /* XXX: if_tzone    Time zone for GMT support (TODO: specify better). */
#define OPT_IDB_FILTER       11    /**< The filter (e.g. "capture only TCP traffic") used to capture traffic.
                                     *     The first byte of the Option Data keeps a code of the filter used
                                     *     (e.g. if this is a libpcap string, or BPF bytecode, and more).
                                     *     More details about this format will be presented in Appendix XXX (TODO).
                                     *     (TODO: better use different options for different fields?
                                     *     e.g. if_filter_pcap, if_filter_bpf, ...) 00 "tcp port 23 and host 10.0.0.5"
                                     */
#define OPT_IDB_OS           12    /**< A UTF-8 string containing the name of the operating system of the
                                     *     machine in which this interface is installed.
                                     *     This can be different from the same information that can be
                                     *     contained by the Section Header Block
                                     *     (Section 3.1 (Section Header Block (mandatory))) because
                                     *     the capture can have been done on a remote machine.
                                     *     "Windows XP SP2" / "openSUSE 10.2"
                                     */
#define OPT_IDB_FCSLEN       13    /**< An integer value that specified the length of the
                                     *     Frame Check Sequence (in bits) for this interface.
                                     *     For link layers whose FCS length can change during time,
                                     *     the Packet Block Flags Word can be used (see Appendix A (Packet Block Flags Word))
                                     */
#define OPT_IDB_TSOFFSET     14    /**< XXX: A 64 bits integer value that specifies an offset (in seconds)
                                     *     that must be added to the timestamp of each packet to obtain
                                     *     the absolute timestamp of a packet. If the option is missing,
                                     *     the timestamps stored in the packet must be considered absolute
                                     *     timestamps. The time zone of the offset can be specified with the
                                     *     option if_tzone. TODO: won't a if_tsoffset_low for fractional
                                     *     second offsets be useful for highly syncronized capture systems?
                                     */
#define OPT_IDB_HARDWARE     15     /**< A UTF-8 string containing the description
                                     *     of the hardware of the device used
                                     *     to capture data.
                                     *     "Broadcom NetXtreme" /
                                     *     "Intel(R) PRO/1000 MT Network Connection" /
                                     *     "NETGEAR WNA1000Mv2 N150 Wireless USB Micro Adapter"
                                     */


#define OPT_NS_DNSNAME       2
#define OPT_NS_DNSIP4ADDR    3
#define OPT_NS_DNSIP6ADDR    4

#define OPT_ISB_STARTTIME    2
#define OPT_ISB_ENDTIME      3
#define OPT_ISB_IFRECV       4
#define OPT_ISB_IFDROP       5
#define OPT_ISB_FILTERACCEPT 6
#define OPT_ISB_OSDROP       7
#define OPT_ISB_USRDELIV     8

struct wtap_block;
typedef struct wtap_block *wtap_block_t;

/*
 * Currently supported blocks; these are not the pcapng block type values
 * for them, they're identifiers used internally.
 */
typedef enum {
    WTAP_BLOCK_NG_SECTION = 0,
    WTAP_BLOCK_IF_DESCR,
    WTAP_BLOCK_NG_NRB,
    WTAP_BLOCK_IF_STATS,
    WTAP_BLOCK_END_OF_LIST
} wtap_block_type_t;

/* Currently supported option types */
typedef enum {
    WTAP_OPTTYPE_UINT8,
    WTAP_OPTTYPE_UINT64,
    WTAP_OPTTYPE_STRING,
    WTAP_OPTTYPE_IPv4,
    WTAP_OPTTYPE_IPv6,
    WTAP_OPTTYPE_CUSTOM
} wtap_opttype_e;

typedef enum {
    WTAP_OPTTYPE_SUCCESS = 0,
    WTAP_OPTTYPE_NO_SUCH_OPTION = -1,
    WTAP_OPTTYPE_NOT_FOUND = -2,
    WTAP_OPTTYPE_TYPE_MISMATCH = -3,
    WTAP_OPTTYPE_NUMBER_MISMATCH = -4,
    WTAP_OPTTYPE_ALREADY_EXISTS = -5
} wtap_opttype_return_val;

struct wtap_opttype_custom
{
    void* data;
    guint size;
};

/*
 * Structure describing a value of an option.
 */
typedef union {
    guint8 uint8val;
    guint64 uint64val;
    guint32 ipv4val;    /* network byte order */
    ws_in6_addr ipv6val;
    char *stringval;
    struct wtap_opttype_custom customval;
} wtap_optval_t;

/*
 * Structure describing an option in a block.
 */
typedef struct {
    guint option_id;     /**< option code for the option */
    wtap_optval_t value; /**< value */
} wtap_option_t;

struct wtap_dumper;

typedef void (*wtap_block_create_func)(wtap_block_t block);
typedef void (*wtap_mand_free_func)(wtap_block_t block);
typedef void (*wtap_mand_copy_func)(wtap_block_t dest_block, wtap_block_t src_block);

typedef void (*wtap_block_foreach_func)(wtap_block_t block, guint option_id, wtap_opttype_e option_type, wtap_optval_t *option, void *user_data);
"""

wtap_opttypes_h_funcs_cdef = """
/** Initialize block types.
 *
 * This is currently just a placeholder as nothing needs to be
 * initialized yet.  Should handle "registration" when code is
 * refactored to do so.
 */
extern void wtap_opttypes_initialize(void);

/** Create a block by type
 *
 * Return a newly allocated block with default options provided
 *
 * @param[in] block_type Block type to be created
 * @return Newly allocated block
 */
extern wtap_block_t wtap_block_create(wtap_block_type_t block_type);

/** Free a block
 *
 * Needs to be called to clean up any allocated block
 *
 * @param[in] block Block to be freed
 */
extern void wtap_block_free(wtap_block_t block);

/** Free an array of blocks
 *
 * Needs to be called to clean up blocks allocated
 * through GArray (for multiple blocks of same type)
 * Includes freeing the GArray
 *
 * @param[in] block_array Array of blocks to be freed
 */
extern void wtap_block_array_free(GArray* block_array);

/** Provide mandatory data of a block
 *
 * @param[in] block Block from which to retrieve mandatory data
 * @return Block mandatory data.  Structure varies based on block type
 */
extern void* wtap_block_get_mandatory_data(wtap_block_t block);

/** Add UINT8 option value to a block
 *
 * @param[in] block Block to which to add the option
 * @param[in] option_id Identifier value for option
 * @param[in] value Value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_add_uint8_option(wtap_block_t block, guint option_id, guint8 value);

/** Set UINT8 option value in a block
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] value New value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_uint8_option_value(wtap_block_t block, guint option_id, guint8 value);

/** Get UINT8 option value from a block
 *
 * @param[in] block Block from which to get the option value
 * @param[in] option_id Identifier value for option
 * @param[out] value Returned value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_get_uint8_option_value(wtap_block_t block, guint option_id, guint8* value) ;

/** Add UINT64 option value to a block
 *
 * @param[in] block Block to which to add the option
 * @param[in] option_id Identifier value for option
 * @param[in] value Value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_add_uint64_option(wtap_block_t block, guint option_id, guint64 value);

/** Set UINT64 option value in a block
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] value New value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_uint64_option_value(wtap_block_t block, guint option_id, guint64 value);

/** Get UINT64 option value from a block
 *
 * @param[in] block Block from which to get the option value
 * @param[in] option_id Identifier value for option
 * @param[out] value Returned value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_get_uint64_option_value(wtap_block_t block, guint option_id, guint64* value) ;

/** Add IPv4 address option value to a block
 *
 * @param[in] block Block to which to add the option
 * @param[in] option_id Identifier value for option
 * @param[in] value Value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_add_ipv4_option(wtap_block_t block, guint option_id, guint32 value);

/** Set IPv4 option value in a block
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] value New value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_ipv4_option_value(wtap_block_t block, guint option_id, guint32 value);

/** Get IPv4 option value from a block
 *
 * @param[in] block Block from which to get the option value
 * @param[in] option_id Identifier value for option
 * @param[out] value Returned value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_get_ipv4_option_value(wtap_block_t block, guint option_id, guint32* value) ;

/** Add IPv6 address option value to a block
 *
 * @param[in] block Block to which to add the option
 * @param[in] option_id Identifier value for option
 * @param[in] value Value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_add_ipv6_option(wtap_block_t block, guint option_id, ws_in6_addr *value);

/** Set IPv6 option value in a block
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] value New value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_ipv6_option_value(wtap_block_t block, guint option_id, ws_in6_addr *value);

/** Get IPv6 option value from a block
 *
 * @param[in] block Block from which to get the option value
 * @param[in] option_id Identifier value for option
 * @param[out] value Returned value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_get_ipv6_option_value(wtap_block_t block, guint option_id, ws_in6_addr* value) ;

/** Add a string option to a block
 *
 * @param[in] block Block to which to add the option
 * @param[in] option_id Identifier value for option
 * @param[in] value Value of option
 * @param[in] value_length Maximum length of string to copy.
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_add_string_option(wtap_block_t block, guint option_id, const char *value, gsize value_length);

/** Add a string option to a block witha printf-formatted string as its value
 *
 * @param[in] block Block to which to add the option
 * @param[in] option_id Identifier value for option
 * @param[in] format printf-like format string
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_add_string_option_format(wtap_block_t block, guint option_id, const char *format, ...);

/** Set string option value in a block
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] value New value of option
 * @param[in] value_length Maximum length of string to copy.
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_string_option_value(wtap_block_t block, guint option_id, const char* value, gsize value_length);

/** Set string option value for nth instance of a particular option in a block
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] idx Instance number of option with that ID
 * @param[in] value New value of option
 * @param[in] value_length Maximum length of string to copy.
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_nth_string_option_value(wtap_block_t block, guint option_id, guint idx, const char* value, gsize value_length);

/** Set string option value in a block to a printf-formatted string
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] format printf-like format string
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_string_option_value_format(wtap_block_t block, guint option_id, const char *format, ...);

/** Get string option value from a block
 *
 * @param[in] block Block from which to get the option value
 * @param[in] option_id Identifier value for option
 * @param[out] value Returned value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_get_string_option_value(wtap_block_t block, guint option_id, char** value) ;

/** Get string option value for nth instance of a particular option in a block
 *
 * @param[in] block Block from which to get the option value
 * @param[in] option_id Identifier value for option
 * @param[in] idx Instance number of option with that ID
 * @param[out] value Returned value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_get_nth_string_option_value(wtap_block_t block, guint option_id, guint idx, char** value) ;

/** Add a "custom" option value to a block
 *
 * @param[in] block Block to which to add the option
 * @param[in] option_id Identifier value for option
 * @param[in] value Value of option
 * @param[in] value_size Size of value
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_add_custom_option(wtap_block_t block, guint option_id, void* value, size_t value_size);

/** Set a "custom" option value in a block
 *
 * @param[in] block Block in which to set the option value
 * @param[in] option_id Identifier value for option
 * @param[in] value New value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_set_custom_option_value(wtap_block_t block, guint option_id, void* value);

/** Get a "custom" option value from a block
 *
 * @param[in] block Block from which to get the option value
 * @param[in] option_id Identifier value for option
 * @param[out] value Returned value of option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_get_custom_option_value(wtap_block_t block, guint option_id, void** value) ;

/** Remove an option from a block
 *
 * @param[in] block Block from which to remove the option
 * @param[in] option_id Identifier value for option
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_remove_option(wtap_block_t block, guint option_id);

/** Remove the nth instance of an option from a block
 *
 * @param[in] block Block from which to remove the option instance
 * @param[in] option_id Identifier value for option
 * @param[in] idx Instance number of option with that ID
 * @return wtap_opttype_return_val - WTAP_OPTTYPE_SUCCESS if successful,
 * error code otherwise
 */
extern wtap_opttype_return_val
wtap_block_remove_nth_option_instance(wtap_block_t block, guint option_id,
                                      guint idx);

/** Copy a block to another.
 *
 * Any options that are in the destination but not the source are not removed.
 * Options that are just in source will be added to destination
 *
 * @param[in] dest_block Block to be copied to
 * @param[in] src_block Block to be copied from
 */
 void wtap_block_copy(wtap_block_t dest_block, wtap_block_t src_block);


 void wtap_block_foreach_option(wtap_block_t block, wtap_block_foreach_func func, void* user_data);

 int wtap_opttype_register_custom_block_type(const char* name, const char* description, wtap_block_create_func create,
                                                wtap_mand_free_func free_mand, wtap_mand_copy_func copy_mand);

/** Cleanup the internal structures
 */
 void wtap_opttypes_cleanup(void);

"""

